#! /usr/bin/env python3
# -*- coding: UTF-8 -*-

"""
gstat_synth.py - GeigerLog tools for the generation of synthetic data

include in programs with:
    include gstat_synth
"""

###############################################################################
#    This file is part of GeigerLog.
#
#    GeigerLog is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GeigerLog is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GeigerLog.  If not, see <http://www.gnu.org/licenses/>.
###############################################################################

__author__          = "ullix"
__copyright__       = "Copyright 2016, 2017, 2018, 2019, 2020, 2021"
__credits__         = [""]
__license__         = "GPL3"

from   gsup_utils       import *



def createSyntheticData():
    """Create synthetic data and save in log format. All times in SECONDS"""

    def shd(d):
        return " ".join("%0.2f" % e for e in d)

    fncname = "createSyntheticData: "
    fprint(header("Create Synthetic Log CSV file"), debug=True)

    default_meanlist = (0.3, 1, 3, 10, 30, 100, 300, 1000, 3000, 10000)

    records     = 100000
    mean        = 30
    cycletime   = 1.0       # in seconds

    # Dialog to input single mean value, or input 0 to use default list
    mean, okPressed = QInputDialog.getDouble(None, "Create Synthetic Log","CPS Mean: Enter desired mean value, or 0 to use this default list for CPS mean:\n{}".format(default_meanlist), 20, min=0, max=1000000, decimals=2)
    if okPressed:
        if mean > 0:    meanlist = (mean,)
        else:           meanlist = default_meanlist
    else:
        fprint("Cancelled")
        return

    setBusyCursor()

    fprint("Create logs for mean(s):", "{}\n".format(shd(meanlist)), debug=True)
    Qt_update()

    for mean in meanlist:
    # get time
        t = np.float64(np.array(list(range(0, records)))) * cycletime
        dprint(fncname + "get time: size: {}, Delta_t: {} sec".format(t.size, (t[1] - t[0])))
        dprint(fncname + "time data: ", np.int_(t))

        # convert time in sec to datetime strings
        strt0 = "2021-01-01 00:00:00"          # nominal default start datetime
        t0    = mpld.datestr2num(strt0)        # time zero
        td    = np.empty(records, dtype='U19') # holds strings of max length 19
        #print("td, len:", td.size, td)
        for i in range(records):
            td[i] = mpld.num2date(t0 + t[i] / 86400.) # clipped after 19 chars by U19 def

        dprint(fncname + "First 3 DateTimes:", td[:3])
        dprint(fncname + "Last  3 DateTimes:", td[-3:])


    # get data - select the distribution you want
        stddev      = np.sqrt(mean)

        sigt, DataSrc     = getWhiteNoisePoisson(records, mean)                             # White Noise from Poisson Distribution
        #sigt, DataSrc    = getWhiteNoiseNormal(records, mean, stddev)                      # White Noise from Normal Distribution
        #sigt, DataSrc    = getSinus(records, t)                                            # Sine (one or more)
        #sigt, DataSrc    = getWhiteNoiseSine(records, t)                                   # White Noise with Sine
        #sigt, DataSrc    = getConstantData(records, t)                                     # constant + breaks
        #sigt, DataSrc    = getAutocorr(records)                                            # Autocorr as time function
        #sigt, DataSrc    = getRectangle(records)                                           # Rectangle of 20 samples @ 3sec = 1min
        #sigt, DataSrc    = getWhiteNoisePoissonAutocorr(records, mean, cycletime)          # Autocorrelated Poisson Noise !! perhaps problem
        #sigt, DataSrc    = getRandomData(records, mean, stddev)                            # Radom Data

        # Mean=17, followed by mean=29
        #~sigt1, DataSrc     = getWhiteNoisePoisson(int(records / 2), mean=17)                    # White Noise from Poisson Distribution
        #~sigt2, DataSrc2    = getWhiteNoisePoisson(int(records / 2), mean=29)                    # White Noise from Poisson Distribution
        #~sigt = np.concatenate((sigt1 , sigt2))
        #~print("len(sigt):", len(sigt))
        #~DataSrc = "WhiteNoisePoisson,mean=17&29CPM"

        # Mean=17, added to mean=29
        #~sigt1, DataSrc     = getWhiteNoisePoisson(int(records), mean=17)                    # White Noise from Poisson Distribution
        #~sigt2, DataSrc2    = getWhiteNoisePoisson(int(records), mean=29)                    # White Noise from Poisson Distribution
        #~sigt = sigt1 + sigt2
        #~print("len(sigt):", len(sigt))
        #~DataSrc = "WhiteNoisePoisson,mean=17+29CPM"


    # write to log file
        path = os.path.join(gglobs.dataPath, DataSrc + ".csv")
        writeFileW(path, "", linefeed = True)
        writeFileA(path, "#HEADER ," + strt0 + ", SYNTHETIC data: " + DataSrc)
        writeFileA(path, "#LOGGING," + strt0 + ", Start: cycle {} sec, device 'SYNTHETIC'".format(cycletime))
        nsigt = np.array(sigt)
        maxprints = 10
        for i in range(records):
            dt          = td[i]     # DateTime
            cps         = nsigt[i]
            cpm         = int(np.nansum(nsigt[max(0, i - 60): i]))
            writestring = " {:7d},{:19s}, {:}, {:}".format(i, dt, cpm, cps)
            writeFileA(path, writestring)
            if i < maxprints or i >= (records - maxprints): print(writestring)

        fprint("Saved to file: {}\n".format(path), debug=True)
        Qt_update()

    setNormalCursor()


def getWhiteNoisePoisson(records, mean):
    """White noise data drawn from Poisson distribution"""

    fncname = "getWhiteNoisePoisson: "
    DataSrc = "WhiteNoisePoisson_CPSmean={}".format(mean)

    x       = np.random.poisson(mean, size=records)

    xtext   = "WhiteNoisePoisson: size:{}, mean={:0.4f}, var={:0.4f}".format(x.size, np.mean(x), np.var(x))
    fprint("Resulting Data:", xtext, debug=True)

    return x, DataSrc


def getWhiteNoiseNormal(records, mean, stddev):
    """
    White noise data drawn from Normal distribution
    """
    DataSrc = "WhiteNoiseNormal_CPSmean={}_std={:3.2f}".format(mean, stddev)

    x       = np.random.normal(mean, stddev, size=records)

    xtext   = "WhiteNoiseNormal: size:{}, mean={:0.3f}, var={:0.3f}, std={:0.3f}".format(x.size, np.mean(x), np.var(x), np.std(x))
    fprint("Resulting Data:", xtext, debug=True)

    return x, DataSrc


def getWhiteNoisePoissonAutocorr(records, mean, cycletime):
    """White noise data drawn from Poisson distribution"""
    # review results before using!!!

    DataSrc = "Autocorrelated Poisson Noise, CPM={}, cycle={}sec".format(mean, cycletime)

    x    = np.random.poisson(mean / 60., size=records * 60 + 60 )
    print("x.size, x.mean, x.var, x.std. :", x.size, np.mean(x), np.var(x), np.std(x))
    print(x)

    x2   = np.zeros(records)
    print("all zeros: x2.size, x2.mean, x2.var, x2.std. :", x2.size, np.mean(x2), np.var(x2), np.std(x2))

    for i in range(0, records):
        j = i * cycletime
        x2[i] = np.sum(x[j : j + 60])

    print("x2.size, x2.mean, x2.var, x2.std. :", x2.size, np.mean(x2), np.var(x2), np.std(x2))

    return x2, DataSrc


def getSinus(records, t):
    """
    sinus data
    All times in seconds
    """

    DataSrc = "Sinus"
    pi2     = 2.0 * np.pi

    x   = 10.0 + 5.0 * np.sin(t * pi2 / (100  * 60 +  0)  ) # periode von 100   min
    x  +=  0.0 + 5.0 * np.sin(t * pi2 / (400  * 60 +  0)  ) # periode von 400   min
    x  +=  0.0 + 5.0 * np.sin(t * pi2 / ( 30  * 60 +  0)  ) # periode von  50   min
    x  +=  0.0 + 5.0 * np.sin(t * pi2 / (  3  * 60 +  4)  ) # periode von   3.x min
    x  +=  0.0 + 5.0 * np.sin(t * pi2 / (  3  * 60 +  9)  ) # periode von   3.y min
    x  +=  0.0 + 5.0 * np.sin(t * pi2 / (  3  * 60 + 13)  ) # periode von   3.z min
    x  +=  0.0 + 5.0 * np.sin(t * pi2 / (  1           )  ) # periode von   1   sec
    x  +=  0.0 + 5.0 * np.sin(t * pi2 / (  0.5         )  ) # periode von   0.5 sec

    return x, DataSrc


def getConstantData(records, t):
    """
    constant data
    All times in seconds
    """

    DataSrc = "Constant + Breaks"
    x       = np.ones(records) * 1000

    for i in range(0, records - 10, 20):
        x[i]     = 500
        x[i + 1] = -10
        x[i + 2] =   1

    return x, DataSrc


def getAutocorr(records):
    """
    The autocorrelation function as time series to use as input
    All times in seconds
    """

    DataSrc = "Autocorrelation as time series - 1st point=1"

    # a small random component
    mean = 0
    std  = 0.01
    x    = np.random.normal(mean, std, size=records)

    #for a 1 sec cycle time the first 60 records will show a linear down curve
    #for i in range (0, 20):
    #    x[i] = 1 - i * 1/20

    x[0] = 1


    return x, DataSrc


def getRectangle(records):
    """
    A rectangle 1 min long with 20 samples @assumed 3sec cycle, otherwise zeros
    """

    DataSrc = "Rectangle_1min"
    x       = np.zeros(records)
    #print x
    for i in range(20): x[i] = 1
    print("x: len:", len(x), x)

    return x, DataSrc


def getWhiteNoiseSine(records, t):
    """
    White noise with Sine data
    """

    DataSrc = "WhiteNoiseNormal + Sine"
    pi2     = 2.0 * np.pi

    mean = 2500
    std  = np.sqrt(mean)
    x    = np.random.normal(mean, std, size=records)
    p1   = (100 * 60 + 1.4) / 60  # 100,0333 min
    print("getWhiteNoiseSinusData: p1:", p1)
    x   +=  40 * std * (np.sin(t * pi2 / p1 ))
    x   +=  10 * std * (np.sin(t * pi2 / 11 ))

    return x, DataSrc


def getRandomData(records, mean, stddev):
    """random data from a uniform range of 0 ... 2*mean"""

    DataSrc = "Random Data_mean={}_postive_uniform_range0...{}".format(mean, 2 * mean)

    x = mean * 2 * np.random.rand(records)
    print(DataSrc, x)

    return x, DataSrc



#! /usr/bin/env python3
# -*- coding: UTF-8 -*-

"""
RadMonPlus support using a MQTT connection
"""

###############################################################################
#    This file is part of GeigerLog.
#
#    GeigerLog is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GeigerLog is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GeigerLog.  If not, see <http://www.gnu.org/licenses/>.
###############################################################################


"""
Installing MQTT on Ubuntu:
see: https://www.digitalocean.com/community/tutorials/how-to-install-and-secure-the-mosquitto-mqtt-messaging-broker-on-ubuntu-16-04
site includes steps to set up secure server!

To install the software:
    sudo apt-get install mosquitto mosquitto-clients

Testing MQTT, assuming mosqitto running on mysite.com:
Subscribe in terminal 1:
    mosquitto_sub -v -h mysite.com -t ambiomon/#
Publish in terminal 2:
    mosquitto_pub -h mysite.com -t ambiomon/temp -m "hello world"
You should see in terminal 1:
    ambiomon/temp hello world
"""


__author__          = "ullix"
__copyright__       = "Copyright 2016, 2017, 2018, 2019, 2020, 2021"
__credits__         = [""]
__license__         = "GPL3"

from   gsup_utils           import *


try:
    # https://pypi.org/project/paho-mqtt/
    import paho.mqtt.client as mqtt
except Exception as e:
    msg  = "\nPython module 'paho-mqtt' was not found\n"
    msg += "In order to run any IoT device like RadMon this module is required\n"
    msg += "Verify that 'paho-mqtt' is installed using GeigerLog tool: 'gtools/GLpipcheck.py'"
    exceptPrint(e, msg)
    edprint("Halting GeigerLog")
    playWav("err")
    sys.exit()


###############################################################################
# BEGIN MQTT Callbacks
###############################################################################

def on_connect(client, userdata, flags, rc):
    """The callback for when the client receives a CONNACK response from the server"""

    # The value of rc indicates success or not:
    strrc = {
                0: "Connection successful",
                1: "Connection refused - incorrect protocol version",
                2: "Connection refused - invalid gglobs.rm_client identifier",
                3: "Connection refused - server unavailable",
                4: "Connection refused - bad username or password",
                5: "Connection refused - not authorised",
                # 6-255: Currently unused.
            }

    if rc < 6:     src = strrc[rc]
    else:          src = "Unexpected Connection Response"
    gglobs.RMconnect = (rc, src)

    if rc == 0:
        gglobs.RMConnection = True
    else:
        gglobs.RMConnection = False

    dprint(">RadMon: on_connect: Client connected:")
    dprint(">        {:20s}: {}".format("with userdata"     , userdata))
    dprint(">        {:20s}: {}".format("with flags"        , flags))
    dprint(">        {:20s}: {}".format("with result"       , src), " (Result Code: {})".format(rc))

    # Subscribing within 'on_connect()' means that if we lose the connection and
    # reconnect then subscriptions will be renewed.
    #gglobs.rm_client.subscribe("$SYS/#") # gives huge amount of traffic! Careful!!!
    gglobs.rm_client.subscribe(gglobs.RMServerFolder + "#", qos=2)


def on_disconnect (client, userdata, rc):
    """When the client has sent the disconnect message it generates an on_disconnect() callback."""

    # The rc parameter indicates the disconnection state.
    # If MQTT_ERR_SUCCESS (0), the callback was called in response to a
    # disconnect() call.
    # If any other value the disconnection was unexpected, such as might be
    # caused by a network error."""

    strrc = {
                0: "MQTT_ERR_SUCCESS",
                1: "Unexpected disconnection",
            }

    if rc == 0:     src = strrc[rc]
    else:           src = strrc[1]
    gglobs.RMdisconnect = (rc, src)

    gglobs.RMConnection = False         # it is disconnected in any case

    dprint(">RadMon: on_disconnect: Client disconnected:")
    dprint(">        {:20s}: {}".format("with userdata"     , userdata))
    dprint(">        {:20s}: {}".format("with result"       , src), "(Result Code: {})".format(rc))


def on_subscribe(client, userdata, mid, granted_qos):
    """Callback when a subscribe has occured"""

    dprint(">RadMon: on_subscribe: Topic subscribed:")
    dprint(">        {:20s}: {}".format("with userdata"     , userdata))
    dprint(">        {:20s}: {}".format("with mid"          , mid))
    dprint(">        {:20s}: {}".format("with granted_qos"  , granted_qos))


def on_message(client, userdata, msg):
    """The callback for when a PUBLISH message is received from the server"""

    # msg: an instance of MQTTMessage.
    # This is a class with members topic, payload, qos, retain.
    # type of msg.payload: <class 'bytes'>
    vprint(">RadMon: on_message: topic: {:20s}, payload: {:10s}, qos: {}, retain: {}".format(msg.topic, str(msg.payload), msg.qos, msg.retain))

    try:
        msg_payload = float(msg.payload) # if msg.payload == '' (empty) then float failes!
    except Exception as e:
        msg = "RadMon: on_message: float(msg.payload) failed, msg.payload ='{}'".format(msg.payload)
        exceptPrint(e, msg)
        msg_payload = gglobs.NAN

    if   msg.topic == gglobs.RMServerFolder + "temp":     gglobs.rm_temp   = msg_payload    # T
    elif msg.topic == gglobs.RMServerFolder + "pressure": gglobs.rm_press  = msg_payload    # P
    elif msg.topic == gglobs.RMServerFolder + "humid":    gglobs.rm_hum    = msg_payload    # H
    elif msg.topic == gglobs.RMServerFolder + "cpm":      gglobs.rm_cpm    = msg_payload    # CPM*
    elif msg.topic == gglobs.RMServerFolder + "rssi":     gglobs.rm_rssi   = msg_payload    # rssi
    #else:  print("msg.topic: {}, msg.payload: {}".format(msg.topic, msg_payload))   # only dose it not used


def on_publish(client, userdata, mid):
    """ Called when a message that was to be sent using the publish() call
    has completed transmission to the broker"""

    #dprint("on_publish: messageID: mid: ", mid)
    pass


###############################################################################
# END MQTT Callbacks
###############################################################################

def getRadMonValues(varlist):
    """Read all RadMon data; return only when complete"""

    fncname = "getRadMonValues: "

    alldata = {}


    dataIncomplete = False
    for vname in varlist:
        if   vname == "T" and gglobs.rm_temp  == None :
            dataIncomplete = True
            break
        elif vname == "P" and gglobs.rm_press == None :
            dataIncomplete = True
            break
        elif vname == "H" and gglobs.rm_hum   == None :
            dataIncomplete = True
            break
        elif vname in ("CPM", "CPM1st", "CPM2nd", "CPM3rd") and gglobs.rm_cpm == None:
            dataIncomplete = True
            break

    if not dataIncomplete:
        for vname in varlist:
            if   vname == "T":
                Temp            = gglobs.rm_temp
                Temp            = scaleVarValues(vname, Temp, gglobs.ValueScale[vname])
                gglobs.rm_temp  = None
                alldata.update({"T": Temp})

            elif vname == "P":
                Press           = gglobs.rm_press
                Press           = scaleVarValues(vname, Press, gglobs.ValueScale[vname])
                gglobs.rm_press = None
                alldata.update({"P": Press})

            elif vname == "H":
                Hum             = gglobs.rm_hum
                Hum             = scaleVarValues(vname, Hum, gglobs.ValueScale[vname])
                gglobs.rm_hum   = None
                alldata.update({"H": Hum})

            elif vname == "X":
                Rssi            = gglobs.rm_rssi
                Rssi            = scaleVarValues(vname, Rssi, gglobs.ValueScale[vname])
                gglobs.rm_rssi  = None
                alldata.update({"X": Rssi})

            elif vname in ("CPM", "CPM1st", "CPM2nd", "CPM3rd"):
                cpm             = gglobs.rm_cpm
                cpm             = scaleVarValues(vname, cpm, gglobs.ValueScale[vname])
                gglobs.rm_cpm   = None
                alldata.update({vname: cpm})

    printLoggedValues(fncname, varlist, alldata)

    return alldata


def getRadMonInfo(extended=False):
    """Info on the Radmon Device"""

    if gglobs.rm_client  == None:   return "No connected device"

    if gglobs.rm_rssi == None:      rssi = "Not available"
    else:                           rssi = str(gglobs.rm_rssi)

    RMInfo = """Connected Device:             '{}'
Configured Variables:         {}
Geiger Tube Sensitivity:      {:0.1f} CPM/(µSv/h) ({:0.4f} µSv/h/CPM)
""".format(\
                                                  gglobs.RMDeviceDetected, \
                                                  gglobs.RMVariables, \
                                                  #~ gglobs.RMSensitivity )
                                                  gglobs.RMSensitivity, 1 / gglobs.RMSensitivity)

    if extended == True:
        RMInfo += """
Connection:                   wireless, MQTT server {}, port:{}
Folder:                       '{}'
Internal collection time:     60 s
Total cycle time (typical):   67 ... 70 s
Wireless signal strength:     {} (rssi)""".format(\
                                                  gglobs.RMServerIP, \
                                                  gglobs.RMServerPort, \
                                                  gglobs.RMServerFolder, \
                                                  rssi)

    return RMInfo


def printRMDevInfo(extended=False):
    """prints basic info on the RadMon device"""

    setBusyCursor()

    txt = "RadMon Device"
    if extended:  txt += " Extended"
    fprint(header(txt))
    fprint("Configured Connection:", "MQTT on {}:{} topic:{}".format(gglobs.RMServerIP, gglobs.RMServerPort, gglobs.RMServerFolder))
    fprint(getRadMonInfo(extended=extended))

    setNormalCursor()


def initRadMon():
    """Initialize the client and set all Callbacks"""

    fncname = "initRadMon: "
    errmsg  = ""

    if not gglobs.RMActivation:
        errmsg = "RadMon device not activated"
        dprint(fncname + errmsg)
        return errmsg

    dprint(fncname + "Initialzing RadMon")
    setDebugIndent(1)

    gglobs.RMDeviceName = "RadMon+"
    gglobs.RMDeviceDetected     = gglobs.RMDeviceName # no difference so far
    gglobs.Devices["RadMon"][0] = gglobs.RMDeviceDetected

    # set configuration
    if gglobs.RMServerIP     == "auto": gglobs.RMServerIP     = "test.mosquitto.org" # before: "mqtt.eclipse.org"
    if gglobs.RMServerPort   == "auto": gglobs.RMServerPort   = 1883
    if gglobs.RMServerFolder == "auto": gglobs.RMServerFolder = "/"
    if gglobs.RMTimeout      == "auto": gglobs.RMTimeout      = 3

# set client
    # Client(client_id="", clean_session=True, userdata=None, protocol=MQTTv311)
    # protocol: Can be either MQTTv31 or MQTTv311
    # on Ubuntu 12.04.5 LTS  with kernel 2.6.32-042stab134.8 the
    # protocol MQTTv311 is NOT working: (does not make a connect)
    # gglobs.rm_client = mqtt.Client(cname, True, None, mqtt.MQTTv311)
    # it is working when protocol MQTTv31 is used
    gglobs.rm_client = mqtt.Client(client_id="", clean_session=True, userdata=None, protocol=mqtt.MQTTv31)

    # set callbacks
    gglobs.rm_client.on_connect       = on_connect
    gglobs.rm_client.on_subscribe     = on_subscribe
    gglobs.rm_client.on_message       = on_message
    gglobs.rm_client.on_publish       = on_publish
    gglobs.rm_client.on_disconnect    = on_disconnect

    # connect
    dprint(fncname + "connect to: RMServerIP: '{}', RMServerPort: {}".format(gglobs.RMServerIP, gglobs.RMServerPort))
    try:
        gglobs.rm_client.connect(gglobs.RMServerIP, port=gglobs.RMServerPort, keepalive=60)
    except Exception as e:
        srcinfo = fncname + "ERROR: Connection to IoT server failed"
        exceptPrint(e, srcinfo)
        gglobs.rm_client     = None
        errmsg += "<br>ERROR: Connection failed using server IP='{}', port={}".format(gglobs.RMServerIP, gglobs.RMServerPort)
        errmsg += "<br>ERROR: '{}'".format(sys.exc_info()[1])
        errmsg += "<br>{} not connected. Is server offline? Verify server IP and server port".format(gglobs.RMDeviceName)

        setDebugIndent(0)
        return errmsg

    #
    # start looping
    #
    ##  # keep as reminder - alternative looping
    ##  gglobs.rm_client.loop_read()
    ##  for i in range(0,6):
    ##      gglobs.rm_client.loop(timeout=.1)
    #gglobs.rm_client.loop_forever()      # indeed forever, it hangs the system
    #                                     # CTRL-C will disconnect the gglobs.rm_client
    #gglobs.rm_client.loop(timeout=10.0)  # returns control after timeout

    gglobs.rm_client.loop_start()         # threaded loop; must be stopped on exit!
    dprint(fncname + "Threaded loop was started")

    # wait for confirmation of connection
    starttime = time.time()
    timeout   = True
    while time.time() < (starttime  + gglobs.RMTimeout):
        #print("gglobs.RMconnect:", time.time() - starttime, gglobs.RMconnect)
        time.sleep(0.05)
        if gglobs.RMConnection:
            #print("gglobs.RMconnect:", time.time() - starttime, gglobs.RMconnect)
            timeout = False
            break

    if timeout:
        # no callback signalling a connection
        dprint(fncname + "RadMon connection timed out ({} sec)".format(gglobs.RMTimeout), debug=True)
        errmsg += "<br>ERROR: Connection to RadMon+ server failed; RadMon+ inactivated"
    else:
        # set only after successful AND confirmed connect!
        if gglobs.RMSensitivity  == "auto":    gglobs.RMSensitivity = 154  # inverse of 0.0065
        if gglobs.RMVariables    == "auto":    gglobs.RMVariables   = "CPM3rd, T, P, H"

        setCalibrations(gglobs.RMVariables, gglobs.RMSensitivity)
        setLoggableVariables("RadMon", gglobs.RMVariables)

    setDebugIndent(0)
    return errmsg


def terminateRadMon():
    """opposit of init ;-)"""

    if gglobs.rm_client == None: return

    fncname = "terminateRadMon: "

    dprint(fncname)
    setDebugIndent(1)

    gglobs.rm_client.loop_stop()
    dprint(fncname + "client.loop was stopped")

    gglobs.rm_client.disconnect()    # output printed in callback
    dprint(fncname + "client was disconnected")

    gglobs.rm_client = None
    dprint(fncname + "client was set to: None")

    # wait for confirmation of dis-connection
    starttime = time.time()
    timeout   = True
    while time.time() < (starttime  + gglobs.RMTimeout):
        #print("gglobs.RMdisconnect:", time.time() - starttime, gglobs.RMdisconnect)
        time.sleep(0.05)
        if not gglobs.RMConnection:
            #print("gglobs.RMdisconnect:", time.time() - starttime, gglobs.RMdisconnect)
            timeout = False
            break

    if timeout:
        dprint("RadMon dis-connection timed out ({} sec)".format(gglobs.RMTimeout), debug=True)
        efprint("<br>ERROR: Dis-Connection from RadMon+ server failed; RadMon+ inactivated")
        setDebugIndent(0)
        retval = ""
    else:
        retval = "RadMon+"

    setDebugIndent(0)



#! /usr/bin/env python3
# -*- coding: UTF-8 -*-

"""
gdev_labjack.py - GeigerLog commands to handle the LabJack U3 with EI1050 probe

NOTE: verfied to work on Linux; it may work on Mac
      Windows needs a different installation of Exodriver
"""

###############################################################################
#    This file is part of GeigerLog.
#
#    GeigerLog is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GeigerLog is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GeigerLog.  If not, see <http://www.gnu.org/licenses/>.
###############################################################################


__author__          = "ullix"
__copyright__       = "Copyright 2016, 2017, 2018, 2019, 2020, 2021"
__credits__         = [""]
__license__         = "GPL3"

"""
For LabJack Python support see:
https://labjack.com/support/software/examples/ud/labjackpython

For threading example see:
https://github.com/labjack/LabJackPython/blob/master/Examples/EI1050/ei1050SampleApp.py

INSTALLATIONS:
To use this device you need the installation of:
- the so called Exodriver
- the Python module: LabJackPython.py
- the Python module: u3.py
- the Python module: ei1050.py

Exodriver -- Linux-only:
The Exodriver is here: https://labjack.com/support/software/installers/exodriver
The latest version is 2.6.0, compatible with Linux kernels from 2.6.28 onwards.
The exodriver is stored as: /usr/local/lib/liblabjackusb.so.

LabJackPython, u3:
LabJackPython is needed in the version 2.0.0 (Jan 2019) or later. Latest version
(Feb 2021) is 2.0.4 from September 9, 2020.
Location: https://labjack.com/support/software/examples/ud/labjackpython
But preferably the modules are installed via pip (includes u3):
    python3 -m pip install -U LabJackPython
On my Python Virtual instllations 'u3.py' is located at: ( X= 6,7,8,9,10)
/home/ullix/geigerlog/vgl3X/lib/python3.X/site-packages/u3.py

ei1050.py:
https://github.com/labjack/LabJackPython/tree/master/Examples/EI1050
example: https://forums.labjack.com/index.php?showtopic=5804


NOTE:
For detection of a busy LabJack see my discussion "Detecting the presence of an
U3 even when in use elsewhere" at: https://labjack.com/comment/6040


###############################################################################
New LabJackSoftware 2021-02-03:

==== Device Mappings =================================================================
The configuration is determined in the configuration file geigerlog.cfg

Device      :  CPM    CPS    CPM1st CPS1st CPM2nd CPS2nd CPM3rd CPS3rd   T   P   H   X
--------------------------------------------------------------------------------------
LabJack     :  -      -      -      -      -      -      -      -        X   -   X   X
Mapping is valid

==== LabJack Device Info Extended ====================================================
Configured Connection:        USB (auto-configuration)
Connected Device:             'LabJack'
Connected Probe (Status):     EI1050 (0)
Connection:                   USB
Configured Variables:         T, H, X
LabJack Software Versions:
--- LabJackPython                : 2.0.4
--- U3                           : unknown
--- EI1050                       : unknown
LabJack Config:
--- BootloaderVersion            : 0.11
--- CIODirection                 : 0
--- CIOState                     : 0
--- CompatibilityOptions         : 0
--- DAC0                         : 0
--- DAC1                         : 0
--- DAC1Enable                   : 0
--- DeviceName                   : U3B
--- EIOAnalog                    : 255
--- EIODirection                 : 0
--- EIOState                     : 0
--- FIOAnalog                    : 255
--- FIODirection                 : 0
--- FIOState                     : 0
--- FirmwareVersion              : 1.54
--- HardwareVersion              : 1.21
--- LocalID                      : 2
--- ProductID                    : 3
--- SerialNumber                 : 320015845
--- TimerClockConfig             : 0
--- TimerClockDivisor            : 256
--- TimerCounterMask             : 0
--- VersionInfo                  : 1
###############################################################################
"""

from   gsup_utils       import *

try:
    import LabJackPython                        # see installation instructions above
    import ei1050                               # dito
    import u3                                   # dito
    #raise Exception
except Exception as e:
    msg  = "\nPython module 'LabJackPython' was not found\n"
    msg += "In order to run a LabJack U3 device this module is required\n"
    msg += "Verify that 'LabJackPython' is installed using GeigerLog tool: 'gtools/GLpipcheck.py'"
    exceptPrint(e, msg)
    edprint("Halting GeigerLog")
    playWav("err")
    sys.exit()


def setLabJackSettings(a, b, c):
    """Set Settings ... at the LabJack Device"""

    global LJdevice, LJprobe, targetQueue, thread

    # There are no settings (at least not for the EI1050 probe)
    pass


def getLabJackValues(varlist):

    global LJdevice, LJprobe, targetQueue, thread

    fncname = "getLabJackValues: "

    alldata = {}

    latestReading   = None
    while not targetQueue.empty():
        try:
            latestReading = targetQueue.get()
        except Exception as e:
            exceptPrint(fncname + str(e), "targetQueue.get() failed")
            latestReading = gglobs.NAN

    if latestReading != None:
        for vname in varlist:
            if   vname == "T":  # from the probe EI1050
                Temp            = latestReading.getTemperature()
                Temp            = round(scaleVarValues(vname, Temp, gglobs.ValueScale[vname]), 2)
                alldata.update({vname: Temp})

            elif vname == "H":  # from the probe EI1050
                Hum             = latestReading.getHumidity()
                Hum             = round(scaleVarValues(vname, Hum, gglobs.ValueScale[vname]), 2)
                alldata.update({vname: Hum})

            elif vname == "X":  # from the U3 internal sensor
                try:
                    u3t         = LJdevice.getTemperature() - 273.15
                except Exception as e:
                    exceptPrint(fncname + str(e), "LJdevice.getTemperature() failed")
                    u3t = gglobs.NAN
                u3t             = round(scaleVarValues(vname, u3t, gglobs.ValueScale[vname]), 2)
                alldata.update({vname: u3t})

    printLoggedValues(fncname, varlist, alldata)

    return alldata


def initLabJack():
    """Initialize LabJack U3 with probe EI1050"""

    global LJdevice, LJprobe, targetQueue, thread

    fncname = "initLabJack: "

    if not gglobs.LJActivation:
        dprint(fncname + "LabJack device not activated")
        return "LabJack device not activated"

    dprint(fncname + "Initialzing LabJack")
    setDebugIndent(1)
    #print("--- initLabJack LabJackPython.listAll(3):", LabJackPython.listAll(3))

    gglobs.LJDeviceName          = "LabJack"
    gglobs.LJDeviceDetected      = gglobs.LJDeviceName
    gglobs.Devices["LabJack"][0] = gglobs.LJDeviceDetected

    # try to get the versions
    try:    gglobs.LJversionLJP     = LabJackPython.__version__ # version of LabJackPython
    except: gglobs.LJversionLJP     = "unknown"

    try:    gglobs.LJversionU3      = u3.__version__            # version of U3
    except: gglobs.LJversionU3      = "undefined"

    try:    gglobs.LJversionEI1050  = EI1050.__version__        # version of EI1050
    except: gglobs.LJversionEI1050  = "undefined"

    try:    LJdevice.close()                                    # important to detach from USB!
    except: pass

    # open the device
    try:
        lju3 = u3.U3()
    except Exception as e:
        stre    = str(e)

        if   "Couldn't open device. Please check that the device you are trying to open is connected." in stre:
            errmsg = "No LabJack device found. Is it connected?"

        elif "Couldn't open device. Device access or claim error" in stre:
            errmsg = "A LabJack device was found, but it may already be in use by a different program."

        elif "name 'u3' is not defined" in stre:
            errmsg = "Failure loading LabJack driver 'u3'. Please, check installation."

        else:
            errmsg = "Connecting to LabJack device gave unexpected error: {}".format(stre)

        srcinfo = fncname + "ERROR: U3 exception: " + errmsg
        exceptPrint(e, srcinfo)

        gglobs.LJConnection     = False

        return errmsg

    targetQueue = queue.Queue()
    thread = ei1050.EI1050Reader(lju3, targetQueue)
    thread.start()

    LJdevice                = lju3
    LJprobe                 = ei1050.EI1050(LJdevice)
    gglobs.LJConnection     = True

    if gglobs.LJVariables  == "auto":   gglobs.LJVariables = "T, H, X"

    setLoggableVariables("LabJack", gglobs.LJVariables)

    setDebugIndent(0)

    return ""


def terminateLabJack():
    """opposit of init ;-)"""

    global LJdevice, LJprobe, targetQueue, thread

    if not gglobs.LJActivation: return

    fncname = "terminateLabJack: "

    try:
        thread.stop()
        thread.join()       # ???
    except Exception as e:
        print(fncname + "Exception: ", e)

    dprint(fncname)

    if gglobs.LJConnection:
        LJdevice.close()        # important to detach from USB  !!!
        gglobs.LJConnection  = False


def printLJDevInfo(extended=False):
    """prints basic info on the LabJack device"""

    setBusyCursor()

    txt = "LabJack Device"
    if extended:  txt += " Extended"
    fprint(header(txt))
    fprint("Configured Connection:", "USB (auto-configuration)")
    fprint(getLabJackInfo(extended=extended))

    setNormalCursor()


def getLabJackInfo(extended = False):
    """Info on the LabJack Device"""

    global LJdevice, LJprobe, targetQueue, thread

    fncname = "getLabJackInfo: "

    if gglobs.LJActivation == False:
        msg = "LabJack device not activated"
        dprint(fncname + msg)
        return msg

    #print("--- getLabJackInfo LabJackPython.listAll(3):", LabJackPython.listAll(3))
    if gglobs.LJConnection  == True:
        LJInfo = """Connected Device:             '{}'
Connected Probe (Status):     {} ({})
Connection:                   USB
Configured Variables:         {}""".format(\
                                           gglobs.LJDeviceName, \
                                           "EI1050", \
                                           LJprobe.getStatus(), \
                                           gglobs.LJVariables)

        if extended == True:
            LJInfo += "\nLabJack Software Versions:\n"
            LJInfo += "--- {:28s} : {}\n".format("LabJackPython", gglobs.LJversionLJP)
            LJInfo += "--- {:28s} : {}\n".format("U3",            gglobs.LJversionU3)
            LJInfo += "--- {:28s} : {}\n".format("EI1050",        gglobs.LJversionEI1050)
            LJInfo += "LabJack Config:"
            for key, value in sorted(LJdevice.configU3().items()):
                LJInfo += "\n--- {:28s} : {}".format(key, value)
    else:
        LJInfo = """No connected device"""

    return LJInfo




#! /usr/bin/env python3
# -*- coding: utf-8 -*-

"""
gdev_simul.py - GeigerLog commands to handle the simul device, which simulates
                counts using a synthetic Poisson number generator
"""

###############################################################################
#    This file is part of GeigerLog.
#
#    GeigerLog is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GeigerLog is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GeigerLog.  If not, see <http://www.gnu.org/licenses/>.
###############################################################################

__author__          = "ullix"
__copyright__       = "Copyright 2016, 2017, 2018, 2019, 2020, 2021"
__credits__         = [""]
__license__         = "GPL3"


from   gsup_utils       import *


"""
from the config file; currently not used

################ NOT IN USE ###################################################
# SimulCounter PREDICTIVE
# Determines whether CPM is let to accumulate or a prediction is made after
# first counts.
#
# Option auto defaults to 'no'
#
# Options:        auto | yes | no
# Default       = auto
#~SimulPredictive   = auto
SimulPredictive   = no


# SimulCounter PREDICTLIMIT
# Sets the count limit which CPM must have reached before a first CPM-prediction
# is given. Before CPM will be reported as NAN.
# Statistical certainty from Std.Dev = Sqrt(N) (valid for Poisson):
# CPM =  10:   ~30%
# CPM =  25:    20%
# CPM = 100:    10%
#
# Option auto defaults to 25
#
# Options:          auto | <any number >= 0 >
# Default         = auto
SimulPredictLimit = auto
###############################################################################
"""



import ntplib


def getNTPDateTime(NTPversion = 4):
    """Call NTP server and return offset in seconds as reported by NTP"""
    # typical offset is in the order of 10 milli second (offset = +/- 0.010

    # latest NTP version is now: 4;     working versions are 2, 3, 4 (not 1!)
    # https://www.eecis.udel.edu/~mills/ntp/html/release.html
    #
    # ntplib: def request(self, host, version=2, port='ntp', timeout=5):
    #
    # NTP document by IETF:       https://tools.ietf.org/html/rfc5905
    # see: page 23, Figure 7 for functions, and ff for details
    #
    # Reference Timestamp:
    # Time when the system clock was last set or corrected, in NTP timestamp format.
    #
    # Origin Timestamp (org):
    # Time at the client when the request departed for the server, in NTP timestamp format.
    #
    # Receive Timestamp (rec):
    # Time at the server when the request arrived from the client, in NTP timestamp format.
    #
    # Transmit Timestamp (xmt):
    # Time at the server when the response left for the client, in NTP timestamp format.
    #
    # Destination Timestamp (dst):
    # Time at the client when the reply arrived from the server, in NTP timestamp format.

    NTP_SERVERS = [ #'pool.ntp.org',
                    #'asia.pool.ntp.org',
                    #~'oceania.pool.ntp.org',
                    #'north-america.pool.ntp.org',
                    #~'south-america.pool.ntp.org',
                    #'europe.pool.ntp.org',
                    'de.pool.ntp.org',
                    #~'0.de.pool.ntp.org',
                    #~'1.de.pool.ntp.org',
                    #~'2.de.pool.ntp.org',
                    #~'3.de.pool.ntp.org',
                    #~'uk.pool.ntp.org',
                    #~'0.uk.pool.ntp.org',
                  ]

    fncname     = "getNTPDateTime: "
    vprint(fncname)
    setDebugIndent(1)
    orig = recv = dest = offset = gglobs.NAN

    client = ntplib.NTPClient()
    for ntpserver in NTP_SERVERS:
        ntps     = []
        try:
            response = client.request(ntpserver, version=NTPversion, timeout=0.2)
        except Exception as e:
            msg = fncname + "NTPversion:{} ".format(NTPversion)
            exceptPrint(e, msg)
            setDebugIndent(0)
            #return gglobs.NAN
            continue

        #print("response.orig_time: ", response.orig_time) # like: 1614266943.3662605

        orig   = response.orig_time     # all times in seconds
        recv   = response.recv_time
        tx     = response.tx_time
        dest   = response.dest_time
        offset = response.offset

        ntps.append( ["orig"     , orig - orig]) # always zero
        ntps.append( ["recv"     , recv - orig])
        ntps.append( ["tx  "     , tx   - orig])
        ntps.append( ["dest"     , dest - orig])
        ntps.append( ["offset"   , offset])
        #ntps.append( ["calc offset"   , ((recv - orig) + (tx - dest)) / 2])

        # vprint for each server
        if gglobs.verbose:
            for  a in ntps: vprint("NTP Version: {}  {:20s}  {:15s}  {:10.3f} ".format(NTPversion, ntpserver, a[0], a[1]))
            print("")

    setDebugIndent(0)

    return offset # offset in seconds


# NTP
###############################################################################
# SimulCounter


def initSimulCounter():
    """Start the simulator"""

    global cps_record, SimulCounterCalls

    fncname ="initSimulCounter: "

    dprint(fncname + "Initialzing SimulCounter")
    setDebugIndent(1)

    errmsg  = "" # what errors can be here?
    gglobs.SimulDeviceName      = "SimulCounter"
    gglobs.SimulDeviceDetected  = gglobs.SimulDeviceName # no difference so far
    gglobs.Devices["Simul"][0]  = gglobs.SimulDeviceDetected

    if gglobs.SimulMean         == "auto": gglobs.SimulMean         = 0.3                       # "background"
    if gglobs.SimulSensitivity  == "auto": gglobs.SimulSensitivity  = 154                       # CPM/(µSv/h), = 0.065 µSv/h/CPM
    if gglobs.SimulVariables    == "auto": gglobs.SimulVariables    = "CPM, CPS"                # non-predictive;
    #if gglobs.SimulPredictive   == "auto": gglobs.SimulPredictive   = False                     # let CPM accumulate
    #if gglobs.SimulPredictLimit == "auto": gglobs.SimulPredictLimit = 25                        # CPM prediction after this count was reached

    setCalibrations(gglobs.SimulVariables, gglobs.SimulSensitivity)
    setLoggableVariables("Simul", gglobs.SimulVariables)

    gglobs.SimulConnection = True

    np.random.seed(int(time.time()))               # seed the random number generator
    cps_record        = np.full(60, gglobs.NAN)    # for storing last 60 sec of CPS values, fill with NANs
    SimulCounterCalls = 0                          # how many times the get data routine was called

    setDebugIndent(0)

    return errmsg


def terminateSimulCounter():
    """Stop the SimulCounter"""

    fncname ="terminateSimulCounter: "
    if not gglobs.SimulConnection: return fncname + "No SimulCounter connection"

    dprint(fncname)
    gglobs.SimulConnection = False


def xxxxxxxx___getSimulCounterValues(varlist):
    """Read all data; return empty dict when not available"""

    global cps_record, SimulCounterCalls

    fncname = "getSimulCounterValues: "
    alldata = {}

    # take as many samples as the logcycle has sec to cover full cps
    size               = int(max(1, gglobs.logcycle))
    simulcps           = np.random.poisson(gglobs.SimulMean, size=size) # generates the "CPS counts"
    cps_record         = np.append(cps_record, simulcps)[-60:]          # append, but take last 60 values only
    SimulCounterCalls += size
    #~print("--------size: ", size, ", simulcps:", simulcps)
    #~print("cps_record: nansum", np.nansum(cps_record), "cps_record:\n", cps_record)

    for vname in varlist:
        if   vname in ("CPM"):
        # CPM non-predictive
            cpm             = int(np.nansum(cps_record))        # must use int or blob will be stored
            cpm             = round(scaleVarValues(vname, cpm, gglobs.ValueScale[vname]), 2)
            alldata.update(  {vname: cpm})

        elif   vname in ("CPM1st"):
        # CPM1st trying the GQ Fast Estimate Time
            cpm             = np.nanmean(cps_record[-3:]) # 1ast 3
            cpm             = int(cpm * 60)
            cpm             = round(scaleVarValues(vname, cpm, gglobs.ValueScale[vname]), 2)
            alldata.update(  {vname: cpm})

        elif vname in ("T"):
        # old: T as (NTP time offset in ms), NTPversion=2
        # T as (NTP time offset in ms), NTPversion=4
            #~t             = getNTPDateTime(NTPversion=2) * 1000
            t             = getNTPDateTime(NTPversion=4) * 1000
            t             = round(scaleVarValues(vname, t, gglobs.ValueScale[vname]), 2)
            alldata.update(  {vname: t})


        elif vname in ("P"):
        # P as (NTP time offset in ms), NTPversion=3
            x             = getNTPDateTime(NTPversion=3) * 1000
            x             = round(scaleVarValues(vname, x, gglobs.ValueScale[vname]), 2)
            alldata.update(  {vname: x})


        elif vname in ("H"):
        # H as (NTP time offset in ms), NTPversion=3
            x             = getNTPDateTime(NTPversion=3) * 1000
            x             = round(scaleVarValues(vname, x, gglobs.ValueScale[vname]), 2)
            alldata.update(  {vname: x})


        elif vname in ("X"):
        # X as (NTP time offset in ms), NTPversion=4
            x             = getNTPDateTime(NTPversion=4) * 1000
            x             = round(scaleVarValues(vname, x, gglobs.ValueScale[vname]), 2)
            alldata.update(  {vname: x})

        elif vname in ("CPS", "CPS1st", "CPS2nd", "CPS3rd"):
        # CPS - predictive not relevant here
            cps             = int(cps_record[-1])       # must use int or blob will be stored
            cps             = round(scaleVarValues(vname, cps, gglobs.ValueScale[vname] ))
            alldata.update(  {vname: cps})

        elif vname in ("CPM2nd", "CPM3rd"):
        # CPM - predictive
            #~limit   4 --> 50% Stddev
            #~limit  11 --> 30% Stddev
            #~limit  25 --> 20% Stddev
            #~limit 100 --> 10% Stddev

            continue # to not use predictive

            cpm             = int(np.nansum(cps_record)) # must use int or 'blob' will be stored in db
            # make the prediction but only while not a full 60 secs counted
            if SimulCounterCalls < 60:
                if cpm <= gglobs.SimulPredictLimit: cpm = gglobs.NAN
                else:                               cpm = int(cpm / SimulCounterCalls * 60)
            else:
                c1mean = np.nanmean(cps_record[:30]) # 1st 30
                c1stdd = np.nanstd (cps_record[:30])
                c2mean = np.nanmean(cps_record[30:]) # last 30
                c2stdd = np.nanstd (cps_record[30:])
                print("Predict: c1mean: ", c1mean, ", c2mean: ", c2mean)
                print("Predict: c1stdd: ", c1stdd, ", c2stdd: ", c2stdd)
                if abs(c1mean - c2mean) > (c1stdd + c2stdd):
                    cpm = int((c2mean * 30) / 30 * 60)

            cpm             = round(scaleVarValues(vname, cpm, gglobs.ValueScale[vname]), 2)
            alldata.update(  {vname: cpm})
            #~#if gglobs.SimulCounterCalls < 70: print("----------------gglobs.SimulCounterCalls   : ", gglobs.SimulCounterCalls,    " CPM: ", cpm)
            #if gglobs.SimulCounterCalls < 70: print("---------------- SimulCounterCalls   : ", SimulCounterCalls,    " CPM: ", cpm)

    printLoggedValues(fncname, varlist, alldata)

    return alldata



def getSimulCounterValues(varlist):
    """Read all data; return empty dict when not available"""

    global cps_record, SimulCounterCalls

    fncname = "getSimulCounterValues: "
    alldata = {}

    # take as many samples as the logcycle has sec to cover full cps
    size               = int(max(1, gglobs.logcycle))
    simulcps           = np.random.poisson(gglobs.SimulMean, size=size) # generates the "CPS counts"
    cps_record         = np.append(cps_record, simulcps)[-60:]          # append, but take last 60 values only
    SimulCounterCalls += size
    #~print("--------size: ", size, ", simulcps:", simulcps)
    #~print("cps_record: nansum", np.nansum(cps_record), "cps_record:\n", cps_record)

    for vname in varlist:
        if   vname in ("CPM", "CPM1st", "CPM2nd", "CPM3rd"):
        # CPM
            cpm             = int(np.nansum(cps_record))        # must use int or blob will be stored
            cpm             = round(scaleVarValues(vname, cpm, gglobs.ValueScale[vname]), 2)
            alldata.update(  {vname: cpm})

        elif vname in ("CPS", "CPS1st", "CPS2nd", "CPS3rd"):
        # CPS
            cps             = int(cps_record[-1])               # must use int or blob will be stored
            cps             = round(scaleVarValues(vname, cps, gglobs.ValueScale[vname] ))
            alldata.update(  {vname: cps})

        elif vname in ("T"):
        # old: T as (NTP time offset in ms), NTPversion=2
        # T as (NTP time offset in ms), NTPversion=4
            #~t             = getNTPDateTime(NTPversion=2) * 1000
            t             = getNTPDateTime(NTPversion=4) * 1000 # in milli-seconds
            t             = round(scaleVarValues(vname, t, gglobs.ValueScale[vname]), 2)
            alldata.update(  {vname: t})

    printLoggedValues(fncname, varlist, alldata)

    return alldata


def getSimulProperties():
    """Set mean and predictive limit"""

    fncname = "getSimulProperties: "
    dprint(fncname)
    setDebugIndent(1)

    lmean = QLabel("CPS Mean\n(0 ... 10000)")
    lmean.setAlignment(Qt.AlignLeft)
    mean = QLineEdit()
    mean.setToolTip('The mean of the Poisson distribution for the CPS variable')
    mean.setText("{:0.3g}".format(gglobs.SimulMean))

    lplimit = QLabel("Prediction Limit\n(not negative)")
    lplimit.setAlignment(Qt.AlignLeft)
    plimit  = QLineEdit()
    plimit.setToolTip('The limit for prediction')
    plimit.setText("{:0.3g}".format(gglobs.SimulPredictLimit))

    graphOptions=QGridLayout()
    graphOptions.addWidget(lmean,           0, 0)
    graphOptions.addWidget(mean,            0, 1)
    #~graphOptions.addWidget(lplimit,         1, 0) # do not show
    #~graphOptions.addWidget(plimit,          1, 1) # do not show

    d = QDialog() # set parent to None to popup in center of screen
    d.setWindowIcon(gglobs.iconGeigerLog)
    d.setFont(gglobs.fontstd)
    d.setWindowTitle("SimulCounter Properties")
    d.setWindowModality(Qt.ApplicationModal)
    #d.setWindowModality(Qt.NonModal)
    #d.setWindowModality(Qt.WindowModal)
    d.setMinimumWidth(300)

    bbox = QDialogButtonBox()
    bbox.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok )
    bbox.accepted.connect(lambda: d.done(100))
    bbox.rejected.connect(lambda: d.done(0))

    layoutV = QVBoxLayout(d)
    layoutV.addLayout(graphOptions)
    layoutV.addWidget(bbox)

    retval = d.exec_()
    #print("reval:", retval)

    if retval == 0:
        dprint(fncname + "Cancelled, Properties unchanged")
        #fprint("\nSet Properties: Cancelled, Properties unchanged")
    else:
    # mean
        mean        = mean.text().replace(",", ".")    #replace any comma with dot
        try:    lm  = round(abs(float(mean)), 1)
        except: lm  = gglobs.SimulMean
        if lm <= 10000: gglobs.SimulMean = lm
        else:           efprint("Illegal entry for mean: ", lm)

    # plimit
        plimit      = plimit.text().replace(",", ".")  #replace any comma with dot
        try:    lp  = round(abs(float(plimit)), 1)
        except: lp  = gglobs.SimulPredictLimit
        gglobs.SimulPredictLimit = lp

        dprint(fncname + "ok, new settings: ", gglobs.SimulMean, " ", gglobs.SimulPredictLimit)

        fprintSimulCounterInfo(extended=False)

    setDebugIndent(0)


def fprintSimulCounterInfo(extended=False):
    """prints basic info on the SimulCounter device"""

    setBusyCursor()

    txt = "SimulCounter Device"
    if extended:  txt += " Extended"
    fprint(header(txt))
    fprint("Configured Connection:", "Poisson Generator")
    fprint(getSimulCounterInfo(extended=extended))

    setNormalCursor()


# only when predictive is used
#~def extendedgetSimulCounterInfo(extended = False):
    #~"""Info on the SimulCounter"""

    #~if not gglobs.SimulConnection:   return "No connected device"

    #~SimulInfo = """Connected Device:             '{}'
#~Configured Variables:         {}
#~Geiger Tube Sensitivity:      {:0.1f} CPM/(µSv/h) ({:0.4f} µSv/h/CPM)
#~Distribution Mean:            {:0.1f} CPS, = {:0.1f} CPM
#~Predictive:                   {}
#~Predictive Limit:             {}
#~"""\
                #~.format(
                        #~gglobs.SimulDeviceName,
                        #~gglobs.SimulVariables,
                        #~gglobs.SimulSensitivity, 1 / gglobs.SimulSensitivity,
                        #~gglobs.SimulMean, gglobs.SimulMean * 60,
                        #~"Yes" if gglobs.SimulPredictive else "No",
                        #~gglobs.SimulPredictLimit,
                       #~)

    #~if extended == True:
        #~SimulInfo += """
#~{}
#~"""\
                #~.format("None")
        #~SimulInfo += "\n"

    #~return SimulInfo


def getSimulCounterInfo(extended = False):
    """Info on the SimulCounter"""

    if not gglobs.SimulConnection:   return "No connected device"

    SimulInfo = """Connected Device:             '{}'
Configured Variables:         {}
Geiger Tube Sensitivity:      {:0.1f} CPM/(µSv/h) ({:0.4f} µSv/h/CPM)
Distribution Mean:            {:0.1f} CPS, = {:0.1f} CPM
"""\
                .format(
                        gglobs.SimulDeviceName,
                        gglobs.SimulVariables,
                        gglobs.SimulSensitivity, 1 / gglobs.SimulSensitivity,
                        gglobs.SimulMean, gglobs.SimulMean * 60,
                       )

    return SimulInfo




#! /usr/bin/env python3
# -*- coding: UTF-8 -*-

"""
I2C module BME280
"""

###############################################################################
#    This file is part of GeigerLog.
#
#    GeigerLog is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GeigerLog is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GeigerLog.  If not, see <http://www.gnu.org/licenses/>.
###############################################################################


__author__          = "ullix"
__copyright__       = "Copyright 2016, 2017, 2018, 2019, 2020, 2021"
__credits__         = [""]
__license__         = "GPL3"

from   gsup_utils       import *

"""
Barometric pressure: (use for altitude correction)
Sea level standard pressure:  1013.25 hPa
Barometric formula: p = p0 * exp(- H / H0), with H0 =7990m
p @ Altitude 85m:  p= 1013.25 * exp(-85/7990) = 1013.25 * 0,98942  = 1002,53

Info on Sensor Module Bluedot:
https://www.bluedot.space/bme280-tsl2591/
Vendor Reference: BME280 + TSL2591, ASIN: B0795WWXX8
"Please note that the BME280 is hardwired to use the I2C Address 0x77. The
alternative Address (0x76) is not available! "
"""

class SensorBME280:
    """Code for the BME280 sensors"""

# =============================================================================
# comparison of results obtained with the different dongles
# all calibration points agree


# Calibration Data for the BlueDot module BME280 + TSL2591
# validation: data identical between the 2 dongles
#
# via ELV dongle
# ELV BME280 TX 11:04:16   [   ] [ 17]  get cal1             == b'S EE 88 S EF 18 P'
# ELV        RX 11:04:16   [ 24] [ 72]                       == b'BE 6E 9A 69 32 00 77 92 CE D6 D0 0B 00 22 A4 FF F9 FF AC 26 0A D8 BD 10 '
# ELV BME280 TX 11:04:16   [   ] [ 17]  get cal2             == b'S EE A1 S EF 01 P'
# ELV        RX 11:04:16   [  1] [  3]                       == b'4B '
# ELV BME280 TX 11:04:16   [   ] [ 17]  get cal3             == b'S EE E1 S EF 07 P'
# ELV        RX 11:04:16   [  7] [ 21]                       == b'6B 01 00 13 2D 03 1E '

# via IOW dongle
# IOW BME280 TX 11:57:37   [  2] [  8]  get cal1             == 02 C2 EE 88 00 00 00 00
# IOW BME280 RX 11:57:37   [ 24] [  8]                       == 02 02 00 00 00 00 00 00 ACK
# IOW BME280 tx 11:57:37   [ 24] [  8]                       == 03 18 EF 00 00 00 00 00
# IOW BME280 RX 11:57:37   [ 24] [  8]                       == 03 06 BE 6E 9A 69 32 00 ok, Bytes received: 6
# IOW BME280 RX 11:57:37   [ 24] [  8]                       == 03 06 77 92 CE D6 D0 0B ok, Bytes received: 12
# IOW BME280 RX 11:57:37   [ 24] [  8]                       == 03 06 00 22 A4 FF F9 FF ok, Bytes received: 18
# IOW BME280 RX 11:57:37   [ 24] [  8]                       == 03 06 AC 26 0A D8 BD 10 ok, Bytes received: 24
# IOW BME280 RX 11:57:37   [ 24] [ 24]      Final            == BE 6E 9A 69 32 00 77 92 CE D6 D0 0B 00 22 A4 FF F9 FF AC 26 0A D8 BD 10
# IOW BME280 TX 11:57:37   [  2] [  8]  get cal2             == 02 C2 EE A1 00 00 00 00
# IOW BME280 RX 11:57:37   [  1] [  8]                       == 02 02 00 00 00 00 00 00 ACK
# IOW BME280 tx 11:57:37   [  1] [  8]                       == 03 01 EF 00 00 00 00 00
# IOW BME280 RX 11:57:37   [  1] [  8]                       == 03 01 4B 00 00 00 00 00 ok, Bytes received: 6
# IOW BME280 RX 11:57:37   [  1] [  1]      Final            == 4B
# IOW BME280 TX 11:57:37   [  2] [  8]  get cal3             == 02 C2 EE E1 00 00 00 00
# IOW BME280 RX 11:57:37   [  7] [  8]                       == 02 02 00 00 00 00 00 00 ACK
# IOW BME280 tx 11:57:37   [  7] [  8]                       == 03 07 EF 00 00 00 00 00
# IOW BME280 RX 11:57:37   [  7] [  8]                       == 03 06 6B 01 00 13 2D 03 ok, Bytes received: 6
# IOW BME280 RX 11:57:37   [  7] [  8]                       == 03 01 1E 01 00 13 2D 03 ok, Bytes received: 12
# IOW BME280 RX 11:57:37   [  7] [  7]      Final            == 6B 01 00 13 2D 03 1E

# ISS BME280  i1 14:40:32   [ 24] [  4]  get cal1            == b'U\xef\x88\x18'           == 55 EF 88 18
# ISS         RX 14:40:32   [ 24] [ 24]                      == b'\xben\x9ai2\x00w\x92\xce\xd6\xd0\x0b\x00"\xa4\xff\xf9\xff\xac&\n\xd8\xbd\x10'
#                                                                                          == BE 6E 9A 69 32 00 77 92 CE D6 D0 0B 00 22 A4 FF F9 FF AC 26 0A D8 BD 10
# ISS BME280  i1 14:40:32   [  1] [  4]  get cal2            == b'U\xef\xa1\x01'           == 55 EF A1 01
# ISS         RX 14:40:32   [  1] [  1]                      == b'K'                       == 4B
# ISS BME280  i1 14:40:32   [  7] [  4]  get cal3            == b'U\xef\xe1\x07'           == 55 EF E1 07
# ISS         RX 14:40:32   [  7] [  7]                      == b'k\x01\x00\x13-\x03\x1e'  == 6B 01 00 13 2D 03 1E

# =============================================================================
# Calling problems with ISS dongle due to non-standard way of reading from device
# preparations

#activating BME280 on ELVdongle +++++++++++++++++++++++++++++++++++++++++++
#ELV BME280  TX 14:12:42   [  9] [  9]  get ID               == b'S EE D0 P'
#ELV         iR 14:12:42   [  9] [  9]                       == b'S EF 01 P'
#ELV BME280  RX 14:12:42   [  1] [  3]                       == b'60 '
#                                                            Found Sensor BME280

#activating BME280 on IOW24-DG +++++++++++++++++++++++++++++++++++++++++++
#IOW BME280  TX 14:04:43   [  2] [  8]  get ID               == 02 C2 EE D0 00 00 00 00
#IOW         RX 14:04:43   [  1] [  8]                       == 02 02 00 00 00 00 00 00 ACK
#IOW         iR 14:04:43   [  1] [  8]                       == 03 01 EF 00 00 00 00 00
#IOW         RX 14:04:43   [  1] [  8]                       == 03 01 60 00 00 00 00 00 ok, Bytes received: 6
#                                                   Answer:  == 60
#                                                            Found Sensor BME280

#activating BME280 on ISSdongle +++++++++++++++++++++++++++++++++++++++++++
# the I2C official way, NOT working !!!
#ISS BME280  TX 15:01:05   [  3] [  3]  get ID               == b'U\xee\xd0'     == 55 EE D0    # send 2 bytes, addr ee + D0
#ISS         iR 15:01:05   [  1] [  3]                       == b'U\xef\x01'     == 55 EF 01    # prep for reading 1 byte from addr EF
#ISS         RX 15:01:06   [  1] [  0]                       == b'' ==                          # Failure, no bytes returned
#
# with ISS specific workaraound
#ISS BME280  TX 14:15:07   [  3] [  3]  get ID               == b'U\xee\xd0'     == 55 EE D0    # send 2 bytes, addr ee + D0
#ISS BME280  i1 14:15:07   [  1] [  4]  get ID               == b'U\xef\xd0\x01' == 55 EF D0 01 # prep for reading 1 byte from addr EF, but RESEND the register D0
#ISS         RX 14:15:07   [  1] [  1]                       == b'`' == 60                      # success
#                                                            Found Sensor BME280

# when not one but two registered need to be addressed, the ELV and IOW work unmodifyied
# but the ISS requires that a different command be used (56 instead of 55)!
# the ISS has no option for 3 byte commands, while the ELV and IOW would work unmodified
# using ELV dongle
#ELV BME280  TX 10:24:11   [ 12] [ 12]  ctrl_meas            == b'S EE F4 D6 P'                         # send 3 bytes, addr ee + F4 D6
#ELV         iR 10:24:11   [  9] [  9]                       == b'S EF 01 P'                            # prep for reading 1 byte from addr EF
#ELV         RX 10:24:11   [  1] [  3]                       == b'D6 '                                  # success

# using ISS dongle
#ISS BME280  TX 10:31:53   [  4] [  4]  ctrl_meas            == b'U\xee\xf4\xd6'     == 55 EE F4 D6     # send 3 bytes, addr ee + F4 D6
#ISS         iR 10:31:53   [  1] [  3]                       == b'U\xef\x01'         == 55 EF 01        # prep for reading 1 byte from addr EF
#ISS         RX 10:31:53   [  1] [  0]                       == b''                  ==                 # Failure, no bytes returned

#ISS BME280  TX 10:27:16   [  4] [  4]  ctrl_meas            == b'U\xee\xf4\xd6'     == 55 EE F4 D6     # send 3 bytes, addr ee + F4 D6
#ISS         ii 10:27:16   [  1] [  5]                       == b'U\xef\xf4\xd6\x01' == 55 EF F4 D6 01  # prep for reading 1 byte from addr EF, resending F4 D6
#ISS         RX 10:27:16   [  1] [  0]                       == b''                  ==                 # Failure, no bytes returned

#ISS BME280  TX 10:47:54   [  4] [  4]  ctrl_meas            == b'U\xee\xf4\xd6'     == 55 EE F4 D6     # send 3 bytes, addr ee + F4 D6
#ISS         ii 10:47:54   [  1] [  5]                       == b'V\xef\xf4\xd6\x01' == 56 EF F4 D6 01  # using CMD=56 (not 55): prep for reading 1 byte from addr EF, resending F4 D6
#ISS         RX 10:47:54   [  1] [  1]                       == b'\xd6'              == D6              # success


    def __init__(self, bme280):
        self.dongle  = bme280["dngl"]    # "ELVdongle", "IOW24-DG", "ISSdongle"
        self.addr    = bme280["addr"]    # 0x76, 0x77
        self.subtype = bme280["type"]    # chip_ID: 0x60
        self.name    = bme280["name"]    # BME280


    def BME280Init(self):
        """Reset, check ID, set reg hum, get calibration, trigger measurement"""

        # soft reset
        data    = [0xE0, 0xB6]
        rbytes  = 0
        answ =        gglobs.elv.ELVaskDongle(self.addr, data, rbytes, name=self.name, info="Soft Reset")

        # check ID
        data    = [0xD0]
        rbytes  = 1
        answ =        gglobs.elv.ELVaskDongle(self.addr, data, rbytes, name=self.name, info="get ID")
        if answ[0] == self.subtype:
            #fprint("Success - Found Sensor BME280")
            pass
        else:
            #~efprint("Failure - Did NOT find Sensor BME280")
            #~return False
            response = "Failure - Did NOT find Sensor BME280"
            return response

        # set ctrl-hum
        # 101 = 5 = oversampling * 16
        data    = [0xf2, 0x05]
        rbytes  = 1
        answ =        gglobs.elv.ELVaskDongle(self.addr, data, rbytes, name=self.name, info="ctrl_hum")

        # Calibration Data calib00...calib25 (0x88 ... 0x9F) 24 values
        data    = [0x88]
        rbytes  = 24
        self.cal1 =        gglobs.elv.ELVaskDongle(self.addr, data, rbytes, name=self.name, info="get cal1")

        # Calibration Data calib26...calib41 (0xA1 ) 1 value
        data    = [0xA1]
        rbytes  = 1
        self.cal2 =        gglobs.elv.ELVaskDongle(self.addr, data, rbytes, name=self.name, info="get cal2")

        # Calibration Data calib26...calib41 (0xe1 ... 0xe7) 7 values
        data    = [0xe1]
        rbytes  = 7
        self.cal3 =        gglobs.elv.ELVaskDongle(self.addr, data, rbytes, name=self.name, info="get cal3")

        # make one measurement to discard (on ISS dongle sometimes measuremnt was wrong)
        self.BME280getTPH()

        #~return True
        return ""


    def BME280getTPH(self):
        """ get one measurement of T, P, H """

        # trigger measurement with: ctrl_meas
        # makes one measurement, then waits for next trigger due to forced mode
        # 0b 101 101 10  = D6 = P oversampling * 16, T oversampling * 16,  forced mode
        data    = [0xf4, 0xd6]
        rbytes  = 1
        answ =        gglobs.elv.ELVaskDongle(self.addr, data, rbytes, name=self.name, info="ctrl_meas")
        #time.sleep(0.05) # appears to be insufficient, occasional faulty result, though it should be plenty:
        #                   BOSCH: t measure,max = 1.25 + [2.3 ⋅ 1] + [2.3 ⋅ 4 + 0.575] + [0] = 13.325 ms
        time.sleep(0.1)

        data    = [0xf7]
        rbytes  = 8
        answ =        gglobs.elv.ELVaskDongle(self.addr, data, rbytes, name=self.name, info="Get data F7...FE", end="")
        #print("BME280getTPH: answ:", answ)

        t_raw, p_raw, h_raw = self.__BME280getRawData(answ)
        t, p, h             = readBME280All(self.cal1, self.cal2, self.cal3, p_raw, t_raw, h_raw)
        wprint("Final Result:  T: {:6.2f}, P: {:6.2f}, H: {:6.2f}".format(t, p, h)) # geht sowieso ins logprint!

        return ("BME280", t, p, h)


    def __BME280getRawData(self, rec):
        """calcs raw press, temp, hum"""

        msb, lsb, xlsb  = rec[0], rec[1], rec[2]
        press           = (msb << 16 | lsb << 8 | xlsb) >> 4

        msb, lsb, xlsb  = rec[3], rec[4], rec[5]
        temp            = (msb << 16 | lsb << 8 | xlsb) >> 4

        msb, lsb        = rec[6], rec[7]
        hum             = (msb <<  8 | lsb )

        return temp, press, hum


    def BME280Reset(self):
        """Soft Reset the BME280 sensor"""

        # 5.4.2 Register 0xE0 “reset”
        # If the value 0xB6 is written to the register, the device is reset
        # using the complete power-on-reset procedure. Writing other values
        # than 0xB6 has no effect. The readout value is always 0x00.
        data    = [0xE0, 0xB6]
        rbytes  = 1
        answ =         gglobs.elv.ELVaskDongle(self.addr, data, rbytes, name=self.name, info="Soft Reset")

        # set ctrl-hum
        # 101 = 5 = oversampling * 16
        data    = [0xf2, 0x05]
        rbytes  = 1
        answ =        gglobs.elv.ELVaskDongle(self.addr, data, rbytes, name=self.name, info="ctrl_hum")



    def BME280getInfo(self):

        info = """{} (Category: {})
- DeviceID:   0x{:02X}
- Address:    0x{:02X}
""".format(self.name, "Temperature, Pressure, Humidity", self.subtype, self.addr)

        return info.split("\n")



    def BME280runAllFunctions(self):
        """ for BME280 sensor """

        # not running until the "answ    = util.askDongle(..." are exchanged for
        #                      " answ    = gglobs.elv.ELVaskDongle(..."
        return

        # Reference Document: "TSL2591 Datasheet - Apr. 2013 - ams163.5"
        # e.g.: https://www.manualshelf.com/manual/adafruit/1980/datasheet-english.html

        # 5.4.1 Register 0xD0 “id”
        # The “id” register contains the chip identification number
        # chip_id[7:0], which is 0x60. This number can be read as soon as the
        # device finished the power-on-reset.
        data    = [0xD0]
        rbytes  = 1
        answ    = util.askDongle(self.dongle, self.addr, data, rbytes, name=self.name, info="get ID")

        # 5.4.2 Register 0xE0 “reset”
        # If the value 0xB6 is written to the register, the device is reset
        # using the complete power-on-reset procedure. Writing other values
        # than 0xB6 has no effect. The readout value is always 0x00.
        data    = [0xE0, 0xB6]
        rbytes  = 1
        answ    = util.askDongle(self.dongle, self.addr, data, rbytes, name=self.name, info="Soft Reset")

        # 5.4.3 Register 0xF2 “ctrl_hum”
        # The “ctrl_hum” register sets the humidity data acquisition options of the device. Changes to this
        # register only become effective after a write operation to “ctrl_meas”.
        # 101 = 5 = oversampling * 16
        data    = [0xf2, 0x05]
        rbytes  = 1
        answ    = util.askDongle(self.dongle, self.addr, data, rbytes, name=self.name, info="ctrl_hum")

        # 5.4.4 Register 0xF3 “status”
        # The “status” register contains two bits which indicate the status of the device.
        # Bit 0 and Bit 3 (when set, something is ongoing
        data    = [0xf3]
        rbytes  = 1
        answ    = util.askDongle(self.dongle, self.addr, data, rbytes, name=self.name, info="status")

        # 5.4.5 Register 0xF4 “ctrl_meas”
        # The “ctrl_meas” register sets the pressure and temperature data acquisition options of the device.
        # The register needs to be written after changing “ctrl_hum” for the changes to become effective.
        # for read out only:
        data    = [0xf4]
        rbytes  = 1
        answ    = util.askDongle(self.dongle, self.addr, data, rbytes, name=self.name, info="ctrl_meas")

        # for write new value:
        # 0b 101 101 10  = D6 = P oversampling * 16, T oversampling * 16,  forced mode
        data    = [0xf4, 0xd6]
        rbytes  = 1
        answ    = util.askDongle(self.dongle, self.addr, data, rbytes, name=self.name, info="ctrl_meas")
        time.sleep(0.1) # wait for measurement complete (never saw a status bit set even w/o sleep)

        # 5.4.6 Register 0xF5 “config”
        # The “config” register sets the rate, filter and interface options of the device. Writes to the “config”
        # register in normal mode may be ignored. In sleep mode writes are not ignored.
        data    = [0xf5]
        rbytes  = 1
        answ    = util.askDongle(self.dongle, self.addr, data, rbytes, name=self.name, info="get config")

        # Read calibration data
        # Calibration Data calib00...calib25 (0x88 ... 0x9F) 24 values
        data    = [0x88]
        rbytes  = 24
        self.cal1    = util.askDongle(self.dongle, self.addr, data, rbytes, name=self.name, info="get cal1")

        # Calibration Data calib26...calib41 (0xA1 ) 1 value
        data    = [0xA1]
        rbytes  = 1
        self.cal2    = util.askDongle(self.dongle, self.addr, data, rbytes, name=self.name, info="get cal2")

        # Calibration Data calib26...calib41 (0xe1 ... 0xe7) 7 values
        data    = [0xe1]
        rbytes  = 7
        self.cal3    = util.askDongle(self.dongle, self.addr, data, rbytes, name=self.name, info="get cal3")


        # when cycling:

        # Since forced mode was chosen, a cycle must begin with ctrl_meas.
        # Must be written to; reading not sufficient
        # 0b 101 101 10  = D6 = P oversampling * 16, T oversampling * 16,  forced mode
        # ctrl_hum is at oversampling * 16 also
        data    = [0xf4, 0xd6]
        rbytes  = 1
        answ    = util.askDongle(self.dongle, self.addr, data, rbytes, name=self.name, info="ctrl_meas")

        # wait until measurement complete
        time.sleep(0.05)

        # get all 8 bytes for T, P, H from F7 onwards
        data    = [0xf7]
        rbytes  = 8
        answ    = util.askDongle(self.dongle, self.addr, data, rbytes, name=self.name, info="data F7...FE")

#--------------------------------------
# code adapted from file bme280.py
# Author : Matt Hawkins
# Date     : 25/07/2016
# http://www.raspberrypi-spy.co.uk/
#--------------------------------------

from ctypes import c_short
from ctypes import c_byte
from ctypes import c_ubyte

def getShort(data, index):
    # return two bytes from data as a signed 16-bit value
    return c_short((data[index+1] << 8) + data[index]).value

def getUShort(data, index):
    # return two bytes from data as an unsigned 16-bit value
    return (data[index+1] << 8) + data[index]

def getChar(data,index):
    # return one byte from data as a signed char
    result = data[index]
    if result > 127:
        result -= 256
    return result

def getUChar(data,index):
    # return one byte from data as an unsigned char
    result =    data[index] & 0xFF
    return result

def readBME280All(cal1, cal2, cal3, pres_raw, t_raw, h_raw):

    # Convert byte data to word values
    dig_T1 = getUShort(cal1, 0)
    dig_T2 = getShort (cal1, 2)
    dig_T3 = getShort (cal1, 4)

    dig_P1 = getUShort(cal1, 6)
    dig_P2 = getShort (cal1, 8)
    dig_P3 = getShort (cal1, 10)
    dig_P4 = getShort (cal1, 12)
    dig_P5 = getShort (cal1, 14)
    dig_P6 = getShort (cal1, 16)
    dig_P7 = getShort (cal1, 18)
    dig_P8 = getShort (cal1, 20)
    dig_P9 = getShort (cal1, 22)

    dig_H1 = getUChar (cal2, 0)

    dig_H2 = getShort (cal3, 0)

    dig_H3 = getUChar (cal3, 2)

    dig_H4 = getChar  (cal3, 3)
    dig_H4 = (dig_H4 << 24) >> 20
    dig_H4 = dig_H4 | (getChar(cal3, 4) & 0x0F)

    dig_H5 = getChar  (cal3, 5)
    dig_H5 = (dig_H5 << 24) >> 20
    dig_H5 = dig_H5 | (getUChar(cal3, 4) >> 4 & 0x0F)

    dig_H6 = getChar  (cal3, 6)

    #Refine temperature
    var1 = ((((t_raw>>3)-(dig_T1<<1)))*(dig_T2)) >> 11
    var2 = (((((t_raw>>4) - (dig_T1)) * ((t_raw>>4) - (dig_T1))) >> 12) * (dig_T3)) >> 14
    t_fine = var1+var2
    temperature = float(((t_fine * 5) + 128) >> 8);

    # Refine pressure and adjust for temperature
    var1 = t_fine / 2 - 64000
    var2 = var1 * var1 * dig_P6 / 32768
    var2 = var2 + var1 * dig_P5 * 2
    var2 = var2 / 4 + dig_P4 * 65536
    var1 = (dig_P3 * var1 * var1 / 524288 + dig_P2 * var1) / 524288
    var1 = (1 + var1 / 32768) * dig_P1
    if var1 == 0:
        pressure=0
    else:
        pressure = 1048576 - pres_raw
        pressure = ((pressure - var2 / 4096.0) * 6250) / var1
        var1 = dig_P9 * pressure * pressure / 2147483648
        var2 = pressure * dig_P8 / 32768
        pressure = pressure + (var1 + var2 + dig_P7) / 16

    # Refine humidity
    humidity = t_fine - 76800
    humidity = (h_raw - (dig_H4 * 64.0 + dig_H5 / 16384.0 * humidity)) * (dig_H2 / 65536.0 * (1.0 + dig_H6 / 67108864.0 * humidity * (1.0 + dig_H3 / 67108864.0 * humidity)))
    humidity = humidity * (1.0 - dig_H1 * humidity / 524288.0)
    if humidity > 100:
        humidity = 100
    elif humidity < 0:
        humidity = 0

    return temperature/100, pressure/100, humidity



#! /usr/bin/env python3
# -*- coding: UTF-8 -*-

"""
gsup_tools.py - GeigerLog tools

include in programs with:
    include gsup_tools
"""

###############################################################################
#    This file is part of GeigerLog.
#
#    GeigerLog is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GeigerLog is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GeigerLog.  If not, see <http://www.gnu.org/licenses/>.
###############################################################################

__author__          = "ullix"
__copyright__       = "Copyright 2016, 2017, 2018, 2019, 2020, 2021"
__credits__         = [""]
__license__         = "GPL3"

from   gsup_utils       import *

import gsup_sql

xt, yt, vline, zero, FitFlag, FitSelector = 0, 0, True, "None", True, "Prop"


def printSuSt():
    """Prints an overview of data from all variables"""

    if gglobs.logTimeSlice is None:
        gglobs.exgg.showStatusMessage("No data available")
        return

    timespan_total = gglobs.logTime[-1] - gglobs.logTime[0]
    timespan_plot  = gglobs.logTimeSlice[-1] - gglobs.logTimeSlice[0]
    size_total     = gglobs.logTime.size
    size_plot      = gglobs.logTimeSlice.size
    cycle_total    = timespan_total / size_total * 86400 # sec
    cycle_plot     = timespan_plot  / size_plot  * 86400 # sec

    fprint(header("Summary Statistics of Variables selected in Plot"))
    fprint("File       = {}".format(gglobs.currentDBPath))
    if os.access(gglobs.currentDBPath , os.R_OK):
        fprint("Filesize   = {:10,.0f} Bytes".format(os.path.getsize(gglobs.currentDBPath)))
    else:
        fprint("Filesize   = File not found!")
    fprint("Records    = {:10,.0f} total,   {:10,.0f} shown in Plot"    .format(size_total, size_plot))
    fprint("Time Span  = {:10,.5f} d total, {:10,.5f} d shown in Plot"  .format(timespan_total, timespan_plot))
    fprint("Avg. Cycle = {:10,.1f} s total, {:10,.1f} s shown in Plot"  .format(cycle_total, cycle_plot))

    fprint("          [Unit]      Avg ±StdDev     Variance          Range            Recs  Last Value")
    for i, vname in enumerate(gglobs.varsBook):
        if gglobs.varcheckedCurrent:
            try:    fprint( gglobs.varlabels[vname])
            except: pass


def getlstats(logTime, logVar, vname):
    """get the text for the QTextBrowser in printStats"""

    vnameFull        = gglobs.varsBook[vname][0]

    logSize          = logTime.size
    logtime_max      = logTime.max()
    logtime_min      = logTime.min()
    logtime_delta    = logtime_max - logtime_min # in days

    records_nonnan   = np.count_nonzero(~np.isnan(logVar))

    logVar_max       = np.nanmax    (logVar)
    logVar_min       = np.nanmin    (logVar)
    logVar_avg       = np.nanmean   (logVar)
    logVar_med       = np.nanmedian (logVar)
    logVar_var       = np.nanvar    (logVar)
    logVar_std       = np.nanstd    (logVar)

    logVar_err       = logVar_std / np.sqrt(logVar.size)
    logVar_sqrt      = np.sqrt(logVar_avg)
    logVar_95        = logVar_std * 1.96   # 95% confidence range

    if vname in ("CPM", "CPM1st", "CPM2nd", "CPM3rd"):
        logVar_text      = "(in units of: {})".format(gglobs.Yunit)

    elif vname in ("CPS",  "CPS1st", "CPS2nd", "CPS3rd"):
        if gglobs.Yunit == "CPM":   unitText = "CPS"
        else:                       unitText = gglobs.Yunit
        logVar_text      = "(in units of: {})".format(unitText)

    else:
        logVar_text      = ""

    ltext = []
    ltext.append("Variable: {} {}\n".format(vnameFull, logVar_text))
    ltext.append("  Recs      = {:10,.0f}  Records".format(records_nonnan))

    if   gglobs.Yunit == "CPM" and (vname in ("CPM", "CPM1st", "CPM2nd", "CPM3rd")):
        ltext.append("  Counts    = {:10,.0f}  Counts calculated as: Average CPM * Log-Duration[min]\n".format(logVar_avg * logtime_delta * 24 * 60))

    elif gglobs.Yunit == "CPM" and (vname in ("CPS", "CPS1st", "CPS2nd", "CPS3rd")):
        ltext.append("  Counts    = {:10,.0f}  Counts calculated as: Average CPS * Log-Duration[sec]\n".format(logVar_avg * logtime_delta * 24 * 60 * 60))

    else:
        ltext.append("")

    #~print("----------------vname: ", vname)
    ltext.append("                       % of avg")
    ltext.append("  Average   ={:8.2f}      100%       Min  ={:8.2f}        Max  ={:8.2f}" .format(logVar_avg,                              logVar_min,          logVar_max)          )
    if logVar_avg != 0:
        ltext.append("  Variance  ={:8.2f} {:8.2f}%"                                          .format(logVar_var,  logVar_var  / logVar_avg * 100))
        ltext.append("  Std.Dev.  ={:8.2f} {:8.2f}%       LoLim={:8.2f}        HiLim={:8.2f}" .format(logVar_std,  logVar_std  / logVar_avg * 100,  logVar_avg-logVar_std,         logVar_avg+logVar_std)  )
        ltext.append("  Sqrt(Avg) ={:8.2f} {:8.2f}%       LoLim={:8.2f}        HiLim={:8.2f}" .format(logVar_sqrt, logVar_sqrt / logVar_avg * 100,  logVar_avg-logVar_sqrt,        logVar_avg+logVar_sqrt) )
        ltext.append("  Std.Err.  ={:8.2f} {:8.2f}%       LoLim={:8.2f}        HiLim={:8.2f}" .format(logVar_err,  logVar_err  / logVar_avg * 100,  logVar_avg-logVar_err,         logVar_avg+logVar_err) )
        ltext.append("  Median    ={:8.2f} {:8.2f}%       P_5% ={:8.2f}        P_95%={:8.2f}" .format(logVar_med,  logVar_med  / logVar_avg * 100,  np.nanpercentile(logVar, 5),   np.nanpercentile(logVar, 95)))
        ltext.append("  95% Conf*)={:8.2f} {:8.2f}%       LoLim={:8.2f}        HiLim={:8.2f}" .format(logVar_95,   logVar_95   / logVar_avg * 100,  logVar_avg-logVar_95,          logVar_avg+logVar_95)   )
    else:
        ltext.append("  Variance  ={:8.2f}  {:>8s}"                                             .format(logVar_var,  "N.A."))
        ltext.append("  Std.Dev.  ={:8.2f}  {:>8s}       LoLim={:8.2f}        HiLim={:8.2f}"   .format(logVar_std,  "N.A."                         ,  logVar_avg-logVar_std,         logVar_avg+logVar_std)  )
        ltext.append("  Sqrt(Avg) ={:8.2f}  {:>8s}       LoLim={:8.2f}        HiLim={:8.2f}"   .format(logVar_sqrt, "N.A."                         ,  logVar_avg-logVar_sqrt,        logVar_avg+logVar_sqrt) )
        ltext.append("  Std.Err.  ={:8.2f}  {:>8s}       LoLim={:8.2f}        HiLim={:8.2f}"   .format(logVar_err,  "N.A."                         ,  logVar_avg-logVar_err,         logVar_avg+logVar_err) )
        ltext.append("  Median    ={:8.2f}  {:>8s}       P_5% ={:8.2f}        P_95%={:8.2f}"   .format(logVar_med,  "N.A."                         ,  np.nanpercentile(logVar, 5),   np.nanpercentile(logVar, 95)))
        ltext.append("  95% Conf*)={:8.2f}  {:>8s}       LoLim={:8.2f}        HiLim={:8.2f}"   .format(logVar_95,   "N.A."                         ,  logVar_avg-logVar_95,          logVar_avg+logVar_95)   )

    return "\n".join(ltext)


def printStats():
    """Printing statistics - will always show only data in current plot"""

    logTime          = gglobs.logTimeSlice           # time data

    if gglobs.logTime is None:
        gglobs.exgg.showStatusMessage("No data available") # when called without a loaded file
        return

    setBusyCursor()

    logSize          = logTime.size
    logtime_size     = logTime.size
    logtime_max      = logTime.max()
    logtime_min      = logTime.min()
    logtime_delta    = logtime_max - logtime_min # in days

    oldest   = (str(mpld.num2date(logtime_min)))[:19]
    youngest = (str(mpld.num2date(logtime_max)))[:19]

    lstats   = QTextBrowser()      # label to hold the description
    lstats.setLineWrapMode(QTextEdit.NoWrap)
    lstats.setTextInteractionFlags(Qt.LinksAccessibleByMouse|Qt.TextSelectableByMouse)

    lstats.append("Data from file: {}\n".format(gglobs.currentDBPath))
    lstats.append("Totals")
    lstats.append("  Filesize  = {:12,.0f} Bytes".format(os.path.getsize(gglobs.currentDBPath)))
    lstats.append("  Records   = {:12,.0f} shown in current plot".format(logSize))
    lstats.append("")
    lstats.append("Legend:   *): Approximately valid for a Poisson Distribution when Average > 10\n")
    lstats.append("="*100)

    lstats.append("Time")
    lstats.append("  Oldest rec    = {}  (time = {:.3f} d)".format(oldest,   0))
    lstats.append("  Youngest rec  = {}  (time = {:.3f} d)".format(youngest, logtime_max - logtime_min))
    lstats.append("  Duration      = {:.1f} s   = {:.3f} m   = {:.3f} h   = {:.3f} d".format(logtime_delta *86400, logtime_delta*1440, logtime_delta *24, logtime_delta))
    lstats.append("  Cycle average = {:0.2f} s".format(logtime_delta *86400/ (logSize -1)))
    lstats.append("="*100)

    for i, vname in enumerate(gglobs.varsBook):
        if gglobs.varcheckedCurrent:
            try:
                logVar  = gglobs.logSliceMod[vname]     # var data
            except:
                continue
            getlstatstext =  getlstats(logTime, logVar, vname)
            lstats.append(getlstatstext)
            lstats.append("="*100)

    lstats.append("First and last few records:\n")
    sql, ruler = gsup_sql.getShowCompactDataSql(gglobs.varcheckedCurrent)
    lstats.append(ruler)
    lstats.append(gglobs.exgg.getExcerptLines(sql, gglobs.currentConn, lmax=7))
    lstats.append(ruler)

    lstats.moveCursor(QTextCursor.Start)

    d = QDialog()
    d.setWindowIcon(gglobs.iconGeigerLog)
    d.setFont(gglobs.fontstd)
    d.setWindowTitle("Statistics on Checked Variables")
    d.setWindowModality(Qt.WindowModal) #Qt.ApplicationModal, Qt.NonModal
    d.setMinimumWidth(1100)
    d.setMinimumHeight(750)

    bbox    = QDialogButtonBox()
    bbox.setStandardButtons(QDialogButtonBox.Ok)
    bbox.accepted.connect(lambda: d.done(0))

    layoutV = QVBoxLayout(d)
    layoutV.addWidget(lstats)
    layoutV.addWidget(bbox)

    setNormalCursor()

    d.exec_()


def selectScatterPlotVars():
    """Selecting the vars for the scatter plot and display and fit options"""

    global xt, yt, vline, zero, FitFlag, FitSelector # to keep the selectios

    if gglobs.logTime is None:              # when called without a loaded file
        gglobs.exgg.showStatusMessage("No data available")
        return

    # X-axis and Y-axis vars
    xlist = QListWidget()
    xlist.setSelectionMode(QAbstractItemView.SingleSelection)
    xlist.addItem("time")
    ylist = QListWidget()
    ylist.setSelectionMode(QAbstractItemView.SingleSelection)
    ylist.addItem("time")
    for i, vname in enumerate(gglobs.varsBook):
        if gglobs.varcheckedCurrent[vname]:
            #~print("gglobs.varcheckedCurrent[vname]: ", vname,  gglobs.varcheckedCurrent[vname])
            xlist.addItem(vname)
            ylist.addItem(vname)

    if xt == 0:
        xlist.setCurrentRow(0)
        ylist.setCurrentRow(0)
    else:
        for i in range(len(xlist)):
            if xlist.item(i).text() == xt: xlist.setCurrentRow(i)
            if ylist.item(i).text() == yt: ylist.setCurrentRow(i)

    # force origin to zero?
    b0 = QRadioButton("Show Optimal ")             # this is default
    b1 = QRadioButton("Show X=0")
    b2 = QRadioButton("Show Y=0")
    b3 = QRadioButton("Show X,Y=0")

    if   zero == "None"      : b0.setChecked(True)
    elif zero == "x"         : b1.setChecked(True)
    elif zero == "y"         : b2.setChecked(True)
    elif zero == "x and y"   : b3.setChecked(True)
    else                     : b0.setChecked(True)

    # draw connecting lines?
    checkb  = QCheckBox("Connecting Line")      # default is yes
    checkb.setChecked(vline)

    layoutT = QHBoxLayout()
    layoutT.addWidget(b3)
    layoutT.addWidget(b1)
    layoutT.addWidget(b2)
    layoutT.addWidget(b0)
    layoutT.addStretch()

    # add polynomial fit?                         default is yes
    checkF = QCheckBox("Add Polynomial Fit of order:")
    checkF.setChecked(FitFlag)

    # selector for order of fit                   default is 1st order = linear
    comboboxF = QComboBox()
    comboboxF.addItems(["Prop", "0", "1", "2", "3", "4", "5", "6", "7"])
    comboboxF.setMaximumWidth(70)
    comboboxF.setToolTip('Select the degree of the polynomial fit')

    #~print("-------------FitSelector: ", FitSelector)
    if FitSelector == "Prop":   comboboxF.setCurrentIndex(0)
    else:                       comboboxF.setCurrentIndex(int(FitSelector) + 1)

    layoutF = QHBoxLayout()
    layoutF.addWidget(checkb)
    layoutF.addWidget(checkF)
    layoutF.addWidget(comboboxF)
    layoutF.addStretch()

    bbox = QDialogButtonBox()
    bbox.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok )
    bbox.accepted.connect(lambda: d.done(100)) # ESC key produces 0 (zero)!
    bbox.rejected.connect(lambda: d.done(-1))

    layoutH0 = QHBoxLayout()
    layoutH0.addWidget(QLabel("X"))
    layoutH0.addWidget(QLabel("Y"))

    layoutH = QHBoxLayout()
    layoutH.addWidget(xlist)
    layoutH.addWidget(ylist)

    d = QDialog()
    d.setWindowIcon(gglobs.iconGeigerLog)
    d.setFont(gglobs.fontstd)
    d.setWindowTitle("Select Variables for Scatter Plot")
    d.setWindowModality(Qt.ApplicationModal)
    d.setMinimumHeight(350)
    #d.setWindowModality(Qt.NonModal)
    #d.setWindowModality(Qt.WindowModal)
    #d.setStyleSheet("QLabel { background-color:#DFDEDD; color:rgb(40,40,40); font-size:30px; font-weight:bold;}")

    layoutV = QVBoxLayout(d)
    layoutV.addLayout(layoutH0)
    layoutV.addLayout(layoutH)
    layoutV.addLayout(layoutT)
    layoutV.addLayout(layoutF)
    layoutV.addWidget(bbox)

    retval = d.exec()
    #print("reval:", retval)
    if retval != 100: return  # ESC key produces 0

    #~print("xlist.currentItem(): ", xlist.currentItem().text())
    #~print("ylist.currentItem(): ", ylist.currentItem().text())
    xt          = xlist.currentItem().text()
    yt          = ylist.currentItem().text()
    vline       = checkb.isChecked()
    FitFlag     = checkF.isChecked()
    FitSelector = comboboxF.currentText()

    if   b0.isChecked():  zero = "None"
    elif b1.isChecked():  zero = "x"
    elif b2.isChecked() : zero = "y"
    elif b3.isChecked() : zero = "x and y"
    else                : zero = "None"

    plotScatter(xt, yt, vline, zero, FitFlag, FitSelector)


def plotScatter(vx, vy, vline, zero, FitFlag, FitSelector):
    """Plotting a Scatter plot"""

    vprint("plotScatter: x:{}, y:{}, vline:{}, zero:{}, FitFlag:{}, FitSelector:{}".format(vx, vy, vline, zero, FitFlag, FitSelector))

    # Plotstyle
    plotstyle        = {'color'             : 'black',
                        'linestyle'         : 'solid' if vline else 'none',
                        'linewidth'         : '0.5',
                        'label'             : 'testlabel',
                        'markeredgecolor'   : 'black',
                        'marker'            : 'o',
                        'markersize'        : '60',
                        'alpha'             : 1,
                        }

    if gglobs.currentDBPath is None: DataSrc = "No Data source!"
    else:                            DataSrc = os.path.basename(gglobs.currentDBPath)

    localvardict = gglobs.varsBook.copy()
    localvardict.update( {'time' : ('time [d]', 'time') } )

    #~print("gglobs.logTimeDiffSlice: ", gglobs.logTimeDiffSlice)
    #~print("gglobs.logSlice: ", gglobs.logSlice)
    #print("vx: ", vx)
    if vx == "time":
        x0 = gglobs.logTimeDiffSlice
    else:
        try:
            x0 = gglobs.logSlice[vx]
        except Exception as e:
            fprint(header("Plot Scatter X"))
            fprint("ERROR in plotScatter: Exception: ", e)
            return

    #print("vy: ", vy)
    if vy == "time":
        y0 = gglobs.logTimeDiffSlice
    else:
        try:
            y0 = gglobs.logSlice[vy]
        except Exception as e:
            fprint(header("Plot Scatter Y"))
            fprint("ERROR in plotScatter: Exception: ", e)
            return

    ymask                   = np.isfinite(y0)      # mask for nan values
    y1                      = y0[ymask]
    x1                      = x0[ymask]

    xmask                   = np.isfinite(x1)      # mask for nan values
    y                       = y1[xmask]
    x                       = x1[xmask]

    if y.size > 0:  plotstyle['markersize'] = float(plotstyle['markersize']) / np.sqrt(y.size)
    else:           plotstyle['markersize'] = 3

    fig2 = plt.figure(facecolor = "#E7F9C9", dpi=gglobs.hidpiScaleMPL)
    gglobs.plotScatterFigNo = plt.gcf().number
    vprint("plotScatter: open figs count: {}, current fig: #{}".format(len(plt.get_fignums()), plt.gcf().number))

    plt.title("Scatter Plot\n", fontsize=12, loc='center')
    subTitleLeft  = DataSrc
    subTitleRight = "Recs:" + str(len(x))
    plt.title(subTitleLeft,  fontsize=10, fontweight='normal', loc = 'left')
    plt.title(subTitleRight, fontsize=10, fontweight='normal', loc = 'right')

    plt.xlabel("x = " + localvardict[vx][0], fontsize=12)
    plt.ylabel("y = " + localvardict[vy][0], fontsize=12)

    plt.grid(True)
    plt.subplots_adjust(hspace=None, wspace=.2 , left=.15, top=0.90, bottom=0.1, right=.97)
    plt.ticklabel_format(useOffset=False)

    # hide the cursor position from showing in the Nav toolbar
    ax1 = plt.gca()
    ax1.format_coord = lambda x, y: ""
    #fig2.canvas.toolbar.set_message = lambda x: "" # suppress all messages???

    # canvas - this is the Canvas Widget that displays the `figure`
    # it takes the `figure` instance as a parameter to __init__
    canvas2 = FigureCanvas(fig2)
    canvas2.setFixedSize(700, 600)
    navtoolbar = NavigationToolbar(canvas2, gglobs.exgg)

    labout  = QTextBrowser() # label to hold the description
    labout.setFont(gglobs.fontstd)
    labout.setMinimumHeight(210)
    txtlines  = "Connecting lines are {}drawn".format("" if vline else "not ")
    txtorigin = "an origin of zero is enforced for {}".format(zero)
    labout.setText("Scatter Plot of y = {} versus x = {}.".format(localvardict[vy][0], localvardict[vx][0]))
    labout.append("{}; {}.".format(txtlines, txtorigin))

    d = QDialog()
    gglobs.plotScatterPointer = d
    d.setWindowIcon(gglobs.iconGeigerLog)
    d.setWindowTitle("Scatter Plot")
    #d.setWindowModality(Qt.ApplicationModal)
    #d.setWindowModality(Qt.NonModal)
    d.setWindowModality(Qt.WindowModal)

    okButton = QPushButton("OK")
    #okButton.setCheckable(True)
    okButton.setAutoDefault(True)
    okButton.clicked.connect(lambda:  d.done(0))

    selectButton = QPushButton("Select")
    #selectButton.setCheckable(True)
    selectButton.setAutoDefault(False)
    selectButton.clicked.connect(lambda:  nextScatterPlot())

    bbox = QDialogButtonBox()
    bbox.addButton(okButton,      QDialogButtonBox.ActionRole)
    bbox.addButton(selectButton,  QDialogButtonBox.ActionRole)

    layoutH = QHBoxLayout()
    layoutH.addWidget(bbox)
    layoutH.addWidget(navtoolbar)
    layoutH.addStretch()
    layoutH.addStretch() # double stretch needed in Py3.6, Py3.9 (others not checked)

    layoutV = QVBoxLayout(d)
    layoutV.addLayout(layoutH)
    layoutV.addWidget(canvas2)
    layoutV.addWidget(labout)

    plt.plot(x, y, **plotstyle) # need to plot first, so xlim, ylim become available!

    xlim = ax1.get_xlim()       # xlim is tuple
    ylim = ax1.get_ylim()

    labout.append("\nPolynomial Coefficients of the Least-Squares-Regression Fit:")
    labout.append("Lowest power first: y = #0 + #1 * x + #2 * x² + #3 * x³ + ...")

    if FitFlag:
        if x.size != 0 and y.size != 0:
            #print("size(x), (y): ", x.size, " ", y.size)

            if FitSelector != "Prop":
                try:
                    pfit = np.polyfit(x, y, int(FitSelector))
                    dprint("plotScatter: pfit: ", pfit)
                    for i,f in enumerate(np.flip(pfit)):
                        labout.append("#{:d} : {: .4g}".format(i, f))

                    p  = np.poly1d(pfit)
                    xs = sorted(x)
                    plt.plot(xs, p(xs), color="red", marker=None, linestyle="solid", linewidth=2)
                except Exception as e:
                    dprint("plotScatter: pfit: Failure with exception: ", e)
                    labout.append("\nFailure to plot. Exception: {}".format(e))
                    playWav("err")

            else:
                def func(x, a):
                    return a * x

                from scipy.optimize import curve_fit
                try:
                    popt, pcov = curve_fit(func, x, y)
                    dprint("plotScatter: curve_fit: popt, pcov, perr: ", popt, pcov, np.sqrt(np.diag(pcov)))
                    labout.append("#0 : {: .4g} (Prop)".format(0))
                    labout.append("#1 : {: .4g}".format(popt[0]))
                    plt.plot(x, func(x, *popt), color="red", marker=None, linestyle="solid", linewidth=2)
                except Exception as e:
                    dprint("plotScatter: curve_fit: Failure with exception: ", e)
                    labout.append("\nFailure to plot. Exception: {}".format(e))
                    playWav("err")
        else:
            dprint("plotScatter: no data in either x or y: x:\n", x, "\ny:\n", y)
            labout.append("\nFailure to plot:")
            if x.size == 0: labout.append("   No data in x")
            if y.size == 0: labout.append("   No data in y")
            playWav("err")

    if   zero == "x" :      plt.xlim(left   = 0)
    elif zero == "y" :      plt.ylim(bottom = 0)
    elif zero == "x and y":
        plt.ylim(bottom = 0)
        plt.xlim(left   = 0)
    elif zero == "None":    pass

    # show window
    fig2.canvas.draw_idle()
    d.exec()
    plt.close(gglobs.plotScatterFigNo)


def nextScatterPlot():
    """closes the dialog and reopens the var selection """

    #print("nextScatter: ")
    gglobs.plotScatterPointer.close()   # closes the dialog
    plt.close(gglobs.plotScatterFigNo)  # closes the figure
    selectScatterPlotVars()


def displayLastValues(self):
    """Displays the last values in big letters"""
    # For the updating see updateDisplayVariableValue()

    # print("+++++++++++++++++++++++++++++++++++++displayLastValues was called")

    if gglobs.logConn == None:          return
    if gglobs.displayLastValsIsOn :   return
    if gglobs.lastValues == None:       return

    gglobs.displayLastValsIsOn = True

    d = QDialog()
    d.setWindowIcon(gglobs.iconGeigerLog)
    d.setFont(gglobs.fontstd)
    d.setWindowTitle("Display Last Log Values")
    d.setWindowModality(Qt.WindowModal)       # can click anywhere, but also can have open many windows
    #d.setWindowModality(Qt.ApplicationModal) # only one window can be open, but can't click anywhere else
    #d.setWindowModality(Qt.NonModal)         # only one window can be open, but can't click anywhere else
    d.setMinimumWidth(420)
    d.setStyleSheet("QLabel {background-color:#DFDEDD; color:rgb(80,80,80);}")

    bbox    = QDialogButtonBox()
    bbox.setStandardButtons(QDialogButtonBox.Ok)
    bbox.accepted.connect(lambda: d.done(0))

    gridlo  = QGridLayout()
    gridlo.setVerticalSpacing(10) # more than default, 5 is about normal?
    layoutV = QVBoxLayout(d)
    lenvn   = len(gglobs.varsBook)
    klabels = [None] * lenvn   # variable names
    dlabels = [None] * lenvn   # device names
    vlabels = [None] * lenvn   # variable values

    t1 = QLabel("Variable")
    t1.setStyleSheet("QLabel {color: #111111; font-size:20px; font-weight:bold;}")
    t2 = QLabel("Device")
    t2.setStyleSheet("QLabel {color: #111111; font-size:20px; font-weight:bold; qproperty-alignment:AlignCenter;}")
    t3 = QLabel("Value")
    t3.setStyleSheet("QLabel {color: #111111; font-size:20px; font-weight:bold; qproperty-alignment:AlignCenter;}")

    gridlo.addWidget(t1,    0, 0)
    gridlo.addWidget(t2,    0, 1)
    gridlo.addWidget(t3,    0, 2)
    gridlo.setColumnStretch(2, 1) # give more space to value col

    for i, vname in enumerate(gglobs.varsBook):
           #print("---i, vname:", i, vname)
            klabels[i] = QLabel(gglobs.varsBook[vname][0])
            klabels[i].setFont((QFont("Sans", 12, weight=QFont.Bold)))
            dlabels[i] = QLabel(" ")

            if not gglobs.varcheckedLog[vname]:
                val = "{:>18s}".format("not mapped")
            else:
                if not np.isnan(gglobs.lastValues[vname]):
                    val = "{:>10.2f}".format(gglobs.lastValues[vname])
                else:
                    val = "{:>10s}".format("  --- ")
            gglobs.exgg.vlabels[i] = QLabel(val)
            gglobs.exgg.vlabels[i].setFont(QFont("Monospace", 22, weight=QFont.Black))
            if gglobs.logging and gglobs.varcheckedLog[vname]:
                gglobs.exgg.vlabels[i].setStyleSheet("QLabel {background-color:#F4D345; color:black; }")
            elif not gglobs.logging and gglobs.varcheckedLog[vname]:
                gglobs.exgg.vlabels[i].setStyleSheet("QLabel {color:darkgray; }")
            else:
                gglobs.exgg.vlabels[i].setStyleSheet("QLabel {color:darkgray; font-size:14px;}")

            for dname in gglobs.Devices:
                #print("dname:", dname, ", vname:", vname)
                if not gglobs.Devices[dname][1] is None:
                    if vname in gglobs.Devices[dname][1]:
                        dlabels[i] = QLabel(dname)
                dlabels[i].setAlignment(Qt.AlignCenter)
                dlabels[i].setStyleSheet("QLabel {color:#111111; font-size:14px;}")

            gridlo.addWidget(klabels[i],             i + 1, 0)
            gridlo.addWidget(dlabels[i],             i + 1, 1)
            gridlo.addWidget(gglobs.exgg.vlabels[i], i + 1, 2)

    layoutV.addLayout(gridlo)

    layoutV.addWidget(bbox)

    d.exec_()
    gglobs.displayLastValsIsOn = False


def printPlotData():
    """Print Data to Notepad as selected in Plot. Data are taken from the plot,
    not from the database!"""

    t0 = gglobs.logTimeSlice

    if t0 is None or len(t0) == 0:
        gglobs.exgg.showStatusMessage("No data available")
        return

    setBusyCursor()

    header = "{:5s}, {:19s}".format("#No", "Datetime")
    for vname in gglobs.varsDefault:
        if gglobs.exgg.varDisplayCheckbox[vname].isChecked():
            header += ", {:>8s}".format(vname)
    fprint(header)

    for i in range(len(t0)):
        vcc = "{:>5d}, {:<19s}".format(i, str(mpld.num2date(t0[i]))[:19])
        for vname in gglobs.varsDefault:
            if gglobs.exgg.varDisplayCheckbox[vname].isChecked():
                vi = gglobs.logSlice[vname][i]
                if    np.isnan(vi): vcc += ", {:>8s}".format("nan")
                else:               vcc += ", {:>8.2f}".format(gglobs.logSlice[vname][i])
        fprint(vcc)
        if gglobs.stopPrinting: break
        Qt_update()

    gglobs.stopPrinting = False

    fprint(header)

    setNormalCursor()


def getLogValues():
    """
    Reads variables CPM, ... etc. from devices, saves it in log file, and
    prints record into LogPad.
    Called by the timer once the timer is started
    """

    start = time.time()

    if not gglobs.logging:      return    # currently not logging
    if gglobs.logConn == None:  return    # no connection defined

    fncname = "getLogValues: "

    import gsup_plot

    gglobs.exgg.dcycl.setStyleSheet("QPushButton {background-color:#F4D345; color:rgb(0,0,0);}")
    Qt_update()

    if gglobs.GMCActivation:        import gdev_gmc         # GMC
    if gglobs.AudioActivation:      import gdev_audio       # AudioCounter
    if gglobs.RMActivation:         import gdev_radmon      # RadMon - then imports "paho.mqtt.client as mqtt"
    if gglobs.AmbioActivation:      import gdev_ambiomon    # AmbioMon
    if gglobs.GSActivation:         import gdev_scout      # GammaScout
    if gglobs.I2CActivation:        import gdev_i2c         # I2C  - then imports dongles and sensor modules
    if gglobs.LJActivation:         import gdev_labjack     # LabJack - then trys to import the LabJack modules
    if gglobs.RaspiActivation:      import gdev_raspi       # Raspi
    if gglobs.SimulActivation:      import gdev_simul       # SimulCounter
    if gglobs.MiniMonActivation:    import gdev_minimon     # MiniMon

    if gglobs.verbose: print()                              # empty line to terminal only
    vprint(fncname + "saving to:", gglobs.logDBPath)
    setDebugIndent(1)

    printLoggedValuesHeader()

    timeJulian, timetag = gsup_sql.DB_getLocaltime() # e.g.: 2458512.928904213, '2019-01-29 10:17:37'
    #~print(fncname + "timetag:", timetag, ",  timeJulian:",timeJulian)

# Reset the logValues to NAN
    logValue = {}                              # logvalue dict
    for vname in gglobs.varsDefault:
        logValue[vname] = gglobs.NAN           # set all to NAN

# get the new values for each device (if active)
    # gglobs.Devices : ("GMC", "Audio", "I2C", "RadMon", "AmbioMon", "LabJack", "Gamma-Scout",  "Raspi", "Simul")
    # e.g.: gglobs.Devices['GMC'][1]    : ['CPM', 'CPS']
    # e.g.: gglobs.Devices['RadMon'][1] : ['T', 'P', 'H', 'R']

    for devname in gglobs.Devices:
        #print("devname:", devname)
        if   devname == "GMC"           and gglobs.GMCConnection:
            logValue.update(gdev_gmc.GMCgetValues(gglobs.Devices[devname][1]))

        elif devname == "Audio"         and gglobs.AudioConnection:
            logValue.update(gdev_audio.getAudioValues(gglobs.Devices[devname][1]))

        elif devname == "RadMon"        and gglobs.RMConnection:
            logValue.update(gdev_radmon.getRadMonValues(gglobs.Devices[devname][1]))

        elif devname == "AmbioMon"      and gglobs.AmbioConnection:
            logValue.update(gdev_ambiomon.getAmbioValues(gglobs.Devices[devname][1]))

        elif devname == "Gamma-Scout"   and gglobs.GSConnection:
            logValue.update(gdev_scout.GSgetValues(gglobs.Devices[devname][1]))

        elif devname == "I2C"           and gglobs.I2CConnection:
            logValue.update(gdev_i2c.getI2CValues(gglobs.Devices[devname][1]))

        elif devname == "LabJack"       and gglobs.LJConnection:
            logValue.update(gdev_labjack.getLabJackValues(gglobs.Devices[devname][1]))

        elif devname == "Raspi"         and gglobs.RaspiConnection:
            logValue.update(gdev_raspi.getRaspiValues(gglobs.Devices[devname][1]))

        elif devname == "Simul"         and gglobs.SimulConnection:
            logValue.update(gdev_simul.getSimulCounterValues(gglobs.Devices[devname][1]))

        elif devname == "MiniMon"       and gglobs.MiniMonConnection:
            logValue.update(gdev_minimon.getMiniMonValues(gglobs.Devices[devname][1]))

# create the Log printstring and print to LogPad
    # Example: 1162 11:43:40   M=143  S=1  M1=  S1=  M2=128.0  S2=3.0  T=25.0  P=983.63  H=24.0  R=18.0
    # timetag: cut-off Date, use time only  '2018-07-14 12:00:52' --> '12:00:52'
    printstring = "{:2.7g} {:8s} " .format(gglobs.LogReadings, timetag[11:])
    for vname in gglobs.varsDefault:
        if gglobs.varcheckedLog[vname]:
            printstring     += " {}=".format(gglobs.varsBook[vname][1])
            #print(fncname + "vname, logValue[vname]: ", vname, logValue[vname])

            if not np.isnan(logValue[vname]):
                printstring += "{:<7.6g}".format(logValue[vname]) # can print 6 digit number as integer
                #~printstring += "{:<6.2f}".format(logValue[vname])
            else:
                printstring += " " * 6
    logPrint(printstring)
    gglobs.lastRecord   = printstring    # needed when a record was snapped
    #print("----------------printstring:", printstring)

# create the database insert and update in-memory data
    datalist     = [None] * (gglobs.datacolsDefault + 1) # (12 + 1) x None
    datalist[0]  = gglobs.LogReadings
    datalist[1]  = "NOW"

    # check for all data being nan
    nanOnly      = True
    for i, vname in enumerate(gglobs.varsBook):
        if not np.isnan(logValue[vname]):
            nanOnly         = False
            datalist[i + 2] = logValue[vname]
    #print("----------------datalist:", datalist)

    # save data, but only if at least one variable is not nan
    if not nanOnly:
        # Write to database
        gsup_sql.DB_insertData(gglobs.logConn, [datalist[0:2] + ["localtime"] + datalist[2:]])

        # update the logDBData array; time is set to matplotlib time
        gglobs.logDBData = np.append(gglobs.logDBData, \
                                                    [[
                                                      timeJulian - gglobs.JULIAN111, \
                                                      logValue["CPM"],    \
                                                      logValue["CPS"],    \
                                                      logValue["CPM1st"], \
                                                      logValue["CPS1st"], \
                                                      logValue["CPM2nd"], \
                                                      logValue["CPS2nd"], \
                                                      logValue["CPM3rd"], \
                                                      logValue["CPS3rd"], \
                                                      logValue["T"],      \
                                                      logValue["P"],      \
                                                      logValue["H"],      \
                                                      logValue["X"],      \
                                                    ]],    \
                                                    axis=0)

# update index (=LogReadings)
    gglobs.LogReadings += 1

# print get values&save duration
    gglobs.LogGetValDur = (time.time() - start)  * 1000
    vprint("{:<25s}{:0.1f} ms".format("get all values and save:", gglobs.LogGetValDur))

# reset cycle button color
    gglobs.exgg.dcycl.setStyleSheet("QPushButton {}")
    Qt_update()

# update the lastValues
    if gglobs.lastValues == None:           # occurs only right after start
        gglobs.lastValues = logValue
    else:
        for vname in gglobs.varsDefault:
            # update only if not nan
            if not np.isnan(logValue[vname]):  gglobs.lastValues[vname] = logValue[vname]
    #print("gglobs.lastValues:",gglobs.lastValues)

# update Value Displays
    gglobs.exgg.updateDisplayVariableValue()   # display lastValues
    Qt_update()

# before graph
    before = time.time()

# update graph, only if graph is the current one!
    if gglobs.activeDataSource == "Log":
        gglobs.currentDBData = gglobs.logDBData       # the data!
        gsup_plot.makePlot()                          # direct plot; slightly quicker than PlotGraph

# before graph: about 20ms with this: Connected: GMC( CPM CPS ); RadMon( T P H R ); Audio( CPM2nd CPS2nd );
# after  graph: about 90...140ms with this: Connected: GMC( CPM CPS ); RadMon( T P H R ); Audio( CPM2nd CPS2nd );
    #~vprint(fncname + "duration up to graph: {:5.1f} ms".format((before      - start)  * 1000))
    #~vprint(fncname + "duration for   graph: {:5.1f} ms".format((time.time() - before) * 1000))
    gglobs.LogPlotDur  = (time.time() - before)  * 1000
    gglobs.LogTotalDur = (time.time() - start)  * 1000

    vprint("{:<25s}{:0.1f} ms".format("Total get & plot:", (time.time() - start)  * 1000))

# relevant only when ESP32 is connected (maybe in conflict with other USB-To Serial devices!)
    #~readSerialConsole()

    setDebugIndent(0)

# end getLogValues##############################################################


def snapLogValue(event):
    """Take a measurement when toolbar icon Snap is clicked"""

    if not gglobs.logging: return

    fncname = "snapLogValue: "
    vprint(fncname)
    setDebugIndent(1)

    getLogValues()

    fprint(header("Snapped Log Values"))
    fprint(gglobs.lastRecord)
    vprint(fncname + gglobs.lastRecord)

    # comment to the DB
    ctype       = "COMMENT"
    cJulianday  = "NOW"
    cinfo       = "Snapped log values: {}".format(gglobs.lastRecord)
    gsup_sql.DB_insertComments(gglobs.logConn, [[ctype, cJulianday, "localtime", cinfo]])

    setDebugIndent(0)



def pushToWeb():
    """Send countrate info to web
    - Presently only the GMC website is supported
    """

    """
    Info on GMCmap:
    from: http://www.gmcmap.com/AutomaticallySubmitData.asp

    Auto submit data URL format:
    http://www.GMCmap.com/log2.asp?AID=UserAccountID&GID=GeigerCounterID &CPM=nCPM&ACPM=nACPM&uSV=nuSV
    At lease one reading data has to be submitted.
        UserAccountID:   user account ID. This ID is assigned once a user registration is completed.
        GeigerCounterID: a global unique ID for each registered Geiger Counter.
        nCPM:  Count Per Minute reading from this Geiger Counter.
        nACPM: Average Count Per Minute reading from this Geiger Counter(optional).
        nuSv:  uSv/h reading from this Geiger Counter(optional).

    Followings are valid data submission examples:
        http://www.GMCmap.com/log2.asp?AID=0230111&GID=0034021&CPM=15&ACPM=13.2&uSV=0.075
        http://www.GMCmap.com/log2.asp?AID=0230111&GID=0034021&CPM=15&ACPM=0&uSV=0
        http://www.GMCmap.com/log2.asp?AID=0230111&GID=0034021&CPM=15&ACPM=0&uSV=0
        http://www.GMCmap.com/log2.asp?AID=0230111&GID=0034021&CPM=15
        http://www.GMCmap.com/log2.asp?AID=0230111&GID=0034021&CPM=15&ACPM=13.2

    The submission result will be returned immediately. Followings are the returned result examples:
        OK.
        Error! User is not found.ERR1.
        Error! Geiger Counter is not found.ERR2.
        Warning! The Geiger Counter location changed, please confirm the location.
     """

    # Button and menu will be greyed out anyway; this can never be reached
    if gglobs.logging == False:
        efprint("Must be logging to update radiation maps")
        return

    # Dialog
    msgbox = QMessageBox()
    msgbox.setWindowIcon(gglobs.iconGeigerLog)
    msgbox.setIcon(QMessageBox.Warning)
    msgbox.setFont(gglobs.fontstd)
    msgbox.setWindowTitle("Updating Radiation World Maps")
    msgbox.setStandardButtons(QMessageBox.Ok)
    msgbox.setDefaultButton(QMessageBox.Ok)
    msgbox.setEscapeButton(QMessageBox.Ok)

    # Exit if no DataSource or His DataSource; need Log DataSource
    if gglobs.activeDataSource == None or gglobs.activeDataSource == "His":
        datatext = "Must show a Log Plot, not His Plot, to update Radiation Maps"
        msgbox.setText(datatext)
        retval = msgbox.exec_()
        return

    cpmdata = gglobs.logSlice["CPM"]
    #print("cpmdata:", cpmdata)
    ymask   = np.isfinite(cpmdata)      # mask for nan values
    cpmdata = cpmdata[ymask]
    #print("cpmdata:", cpmdata)

    lendata = len(cpmdata)
    CPM     = np.mean(cpmdata)
    ACPM    = CPM
    uSV     = CPM / gglobs.calibration1st

    #~# Defintions valid for cfgKeyHigh[key][0] (0=zero)
    #~# holding the definitions of the geigerlog,cfg file!
    #~data         = {}
    #~data['AID']  = cfgKeyHigh["UserID"][0]
    #~data['GID']  = cfgKeyHigh["CounterID"][0]
    #~data['CPM']  = "{:3.1f}".format(CPM)
    #~data['ACPM'] = data['CPM']
    #~data['uSV']  = "{:3.2f}".format(uSV)
    #~gmcmapURL    = cfgKeyHigh["Website"][0] + "/" + cfgKeyHigh["URL"][0]  + '?' + urllib.parse.urlencode(data)

    # Defintions valid for cfgKeyHigh[key][0] (0=zero)
    # holding the definitions of the geigerlog,cfg file!
    data         = {}
    data['AID']  = gglobs.GMCmapUserID
    data['GID']  = gglobs.GMCmapCounterID
    data['CPM']  = "{:3.1f}".format(CPM)
    data['ACPM'] = data['CPM']
    data['uSV']  = "{:3.2f}".format(uSV)
    gmcmapURL    = gglobs.GMCmapWebsite + "/" + gglobs.GMCmapURL + '?' + urllib.parse.urlencode(data)

    strdata = ""
    mapform = "   {:11s}: {}\n"
    strdata += mapform.format("CPM",        data['CPM'])
    strdata += mapform.format("ACPM",       data['ACPM'])
    strdata += mapform.format("uSV",        data['uSV'])
    strdata += mapform.format("UserID",     data['AID'])
    strdata += mapform.format("CounterID",  data['GID'])
    strdata  = strdata[:-1] # remove last linefeed

    # Dialog Confirm Sending
    #~datatext  = "Calling server: " + cfgKeyHigh["Website"][0] + "/" + cfgKeyHigh["URL"][0]
    datatext  = "Calling server: " + gglobs.GMCmapWebsite + "/" + gglobs.GMCmapURL
    datatext += "\nwith these data based on {} datapoints:\n\n".format(lendata)
    datatext += strdata + "\n\nPlease confirm with OK, or Cancel"
    msgbox.setText(datatext)
    msgbox.setStandardButtons(QMessageBox.Ok|QMessageBox.Cancel)
    msgbox.setDefaultButton(QMessageBox.Ok)
    msgbox.setEscapeButton(QMessageBox.Cancel)
    retval = msgbox.exec_()
    if retval != 1024:   return

    msg = "Updating Radiation World Maps"
    fprint(header(msg))
    dprint(msg + " - " +  gmcmapURL)
    setDebugIndent(1)

    # dprint the data which go to web
    for a in data:  dprint("{:5s}: {}".format(a, data[a]))

    try:
        with urllib.request.urlopen(gmcmapURL) as response:
            answer = response.read()
        dprint("Server Response: ", answer)
    except Exception as e:
        answer  = b"Bad URL"
        srcinfo = "Bad URL: " + gmcmapURL
        exceptPrint("pushToWeb: " + str(e), srcinfo)

    # possible gmcmap server responses:
    #   on proper credentials:                    b'\r\n<!--  sendmail.asp-->\r\n\r\nWarrning! Please update/confirm your location.<BR>OK.ERR0'
    #   on wrong credentials:                     b'\r\n<!--  sendmail.asp-->\r\n\r\nError! User not found.ERR1.'
    #   on proper userid but wrong counterid:     b'\r\n<!--  sendmail.asp-->\r\n\r\nError! Geiger Counter not found.ERR2.'
    #   something wrong in the data part of URL:  b'\r\n<!--  sendmail.asp-->\r\n\r\nError! Data Error!(CPM)ERR4.'

    # ERR0 = Ok
    if   b"ERR0" in answer:
        fprint("Successfully updated Radiation World Maps with these data:")
        fprint(strdata)
        fprint("Website response:", answer.decode('UTF-8'))

    # ERR1 or ERR2 - wrong userid  or wrong counterid
    elif b"ERR1" in answer or b"ERR2" in answer :
        efprint("Failure updating Radiation World Maps.")
        qefprint("Website response: ", answer.decode('UTF-8'))

    # misformed URL - wrong entry into config?
    elif b"Bad URL" in answer:
        efprint("Failure updating Radiation World Maps.")
        qefprint(" ERROR: ", "Bad URL: " + gmcmapURL)

    # other errors
    else:
        efprint("Unexpected response updating Radiation World Maps.")
        qefprint("Website response: ", answer.decode('UTF-8'))


    setDebugIndent(0)



#! /usr/bin/env python3
# -*- coding: UTF-8 -*-

"""
I2C module TSL2591 (Light Sensor, Visible + Infrared)
"""

###############################################################################
#    This file is part of GeigerLog.
#
#    GeigerLog is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GeigerLog is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GeigerLog.  If not, see <http://www.gnu.org/licenses/>.
###############################################################################

__author__          = "ullix"
__copyright__       = "Copyright 2016, 2017, 2018, 2019, 2020, 2021"
__credits__         = [""]
__license__         = "GPL3"

from   gsup_utils       import *

# Document: "TSL2591 Datasheet - Apr. 2013 - ams163.5"
# e.g.: https://www.manualshelf.com/manual/adafruit/1980/datasheet-english.html

# Sensor Bluedot:
# Herstellerreferenz: BME280 + TSL2591, ASIN: B0795WWXX8
# Light sensor: address 0x29"

#activating TSL2591 on ELVdongle +++++++++++++++++++++++++++++++++++++++++++
#ELV TSL2591 TX 14:12:43   [  9] [  9]  get ID               == b'S 52 B2 P'
#ELV         iR 14:12:43   [  9] [  9]                       == b'S 53 01 P'
#ELV TSL2591 RX 14:12:43   [  1] [  3]                       == b'50 '
#                                                            Found Sensor TSL2591

#activating TSL2591 on IOW24-DG +++++++++++++++++++++++++++++++++++++++++++
#IOW TSL2591 TX 14:04:43   [  2] [  8]  get ID               == 02 C2 52 B2 00 00 00 00
#IOW         RX 14:04:43   [  1] [  8]                       == 02 02 00 00 00 00 00 00 ACK
#IOW         iR 14:04:43   [  1] [  8]                       == 03 01 53 00 00 00 00 00
#IOW         RX 14:04:43   [  1] [  8]                       == 03 01 50 00 00 00 00 00 ok, Bytes received: 6
#                                                   Answer:  == 50
#                                                            Found Sensor TSL2591

#activating TSL2591 on ISSdongle +++++++++++++++++++++++++++++++++++++++++++
#ISS TSL2591 TX 13:51:03   [  3] [  3]  get ID               == b'UR\xb2' == 55 52 B2
#ISS TSL2591 i1 13:51:03   [  1] [  4]  get ID               == b'US\xb2\x01' == 55 53 B2 01
#ISS         RX 13:51:03   [  1] [  1]                       == b'P' == 50
#                                                            Found Sensor TSL2591


class SensorTSL2591:
    """Code for the TSL2591 sensors"""

    PID         = 0x00      # acc to document, page 16

    # CMD Register = 0b1 01 0 0000  = 0xA0 is: CMD + Normal operation
    CMD         = 0xA0

    # Gain and Integration time determine sensitivity and quality of measurement
    # Gain, doc page 6
    #         Name   Factor
    #AGAIN  = Low    1
    #AGAIN  = Med    25      1
    #AGAIN  = High   428     17.12       1
    #AGAIN  = Max    9876    395.04      23.07
    #
    # Integration time, doc page 13
    #ATIME  = 100 ms
    #ATIME  = 200 ms
    #ATIME  = 300 ms
    #ATIME  = 400 ms
    #ATIME  = 500 ms
    #ATIME  = 600 ms

    integration_time = ["100ms", "200ms", "300ms", "400ms", "500ms", "600ms"]

    #                Name:  FieldVal,  Factor
    sensorgain  = { "Low":  (0b00,     1),
                    "Med":  (0b01,     25),
                    "High": (0b10,     428),
                    "Max":  (0b11,     9876),
                  }

    #                Name:  FieldVal,  ms
    sensorint   = { "100ms":(0b000,    100),
                    "200ms":(0b001,    200),
                    "300ms":(0b010,    300),
                    "400ms":(0b011,    400),
                    "500ms":(0b100,    500),
                    "600ms":(0b101,    600),
                  }


    def __init__(self, TSL2591):

        self.dongle  = TSL2591["dngl"]    # "ELVdongle", "IOW24-DG", "ISSdongle"
        self.addr    = TSL2591["addr"]    # 0x29
        self.subtype = TSL2591["type"]    # Device ID: 0x50
        self.name    = TSL2591["name"]    # TSL2591


    def TSL2591Init(self):
        """check ID, check PID, Reset, enable measurement"""

    # Get Device Identification = 0x50  (= as subtype)
        # ID Register (0x12) (Bit 7:0)
        data    = [self.CMD + 0x12]
        rbytes  = 1
        answ        = gglobs.elv.ELVaskDongle(self.addr, data, rbytes, name=self.name, info="get ID")
        if answ[0] == self.subtype:
            #fprint("Found Sensor TSL2591")
            pass
        else:
            efprint("Did NOT find Sensor TSL2591")
            return False

    # Get package identification (PID)
        # PID Register (0x11) (Bit 5:4) (2 bits only!)
        data    = [self.CMD + 0x11]
        rbytes  = 1
        answ        = gglobs.elv.ELVaskDongle(self.addr, data, rbytes, name=self.name, info="PID read (Bits5:4)")

        pid     = answ[0] & 0b00110000
        if pid == self.PID:
            #fprint("Package Identification 0b{:02b} confirmed".format(pid))
            pass
        else:
            efprint("Package Identification 0b{:02b} not as expected".format(pid))

    # System Reset
        # Control Register (0x01)
        #
        # Important: the System Reset will NOT return an ACK! (observed on both ELV and IOW)
        #
        data    = [self.CMD + 0x01, 0x80] # System Reset, AGAIN=00, ATIME=000
        rbytes  = 1
        #answ        = gglobs.elv.ELVaskDongle(self.addr, data, rbytes, name=self.name, info="System Reset")

    # Enable measurement
        # Enable Register (0x00)
        # Power the device on, enable measurements
        # Bit#0 = 1: Power ON
        # Bit#1 = 1: ALS Enable
        # Register: 0b 0000 00 11  =0x03 :
        data    = [self.CMD + 0x00, 0x03]
        rbytes  = 1
        answ        = gglobs.elv.ELVaskDongle(self.addr, data, rbytes, name=self.name, info="Enable ALS+PON")

        return True


    def TSL2591getLumAuto(self, integration_time="500ms"):
        """get Lum. 1st run fast with 100ms inttime, then with desired inttime"""

        if integration_time in self.integration_time:
            finalatime = integration_time
        else:
            finalatime = "500ms"

        selector  = ("Low",      # Factor = 1
                     "Med",      # Factor = 25
                     "High",     # Factor = 428
                     "Max",      # Factor = 9876
                    )
        selindex  = 1            # start with Med; best chance for success in fewest cycles

        again     = selector[selindex]
        atime     = "100ms"      # make first guess fast; the use final, slow inttime
        firstRun  = True
        breakflag = False

        while True:
            #get a lum value
            vis, ir, visraw, irraw, gainFct, inttime = self.TSL2591getLum(gain=again, intgrl=atime)

            wprint("prelim result: Vis: {:3.3g},   IR: {:3.3g}, Gain: {:4d},   RAW: Vis: {},   IR: {}, IntTime:{}"\
                                  .format(vis, ir, gainFct, visraw, irraw, inttime ))

            # lower limit for autoscale must be <= min(2600, 3800, 2800)
            # chosen is 2500
            #        Name   Factor
            #                                             104
            #AGAIN = Low    1                            2600     152
            #AGAIN = Med    25      1                   65000    3800     163
            #AGAIN = High   428     17.12       1               65000    2800
            #AGAIN = Max    9876    395.04      23,07                   65000

            lastselindex = selindex
            testraw = visraw * 600 / inttime

            if testraw > 65000:                         # too much light
                selindex += -1                          # one step down
                if selindex < 0: breakflag = True       # reached the bottom?
                wprint("Autodecrease Gain 1 step")

            elif testraw < 152:                         # allows two step up:
                selindex += 2                           # one step up
                if selindex > 3: breakflag = True       # broke the ceiling?
                wprint("Autoincrease Gain 2 steps")

            elif testraw < 2500:                        # allows one step up
                selindex += 1                           # one step up
                if selindex > 3: breakflag = True       # broke the ceiling?
                wprint("Autoincrease Gain 1 step")
            else:
                breakflag = True

            if breakflag:
                if firstRun:
                    breakflag = False
                else:
                    break                               # best value possible

            firstRun    = False
            again       = selector[selindex]            # set new gain
            atime       = finalatime                    # set best resolution for final run

        wprint("Final result:  Vis: {:3.3g},   IR: {:3.3g}, Gain: {:4d},   RAW: Vis: {},   IR: {}, IntTime:{}"\
                               .format(vis, ir, gainFct, visraw, irraw, inttime ))

        return ("TSL2591", vis, ir, visraw, irraw, gainFct, inttime)


    def TSL2591getLum(self, gain = 'Low', intgrl = "100ms"):

        #print("-------------------------TSL2591getLum - Begin")

        gainFV  = self.sensorgain[gain][0]   # Field Value
        gainFct = self.sensorgain[gain][1]   # Gain Factor

        intFV   = self.sensorint[intgrl][0]  # Field Value
        intTime = self.sensorint[intgrl][1]  # integration time in ms
        intFct  = intTime / 100              # Gain Factor by integration time

        # Control Register (0x01) - Setting Gain Mode and Integration Time
        data    = [self.CMD + 0x01, gainFV << 4 | intFV ]
        rbytes  = 1
        answ        = gglobs.elv.ELVaskDongle(self.addr, data, rbytes, name=self.name, info="Gain:{}, Int:{} ms".format(gainFct, intTime))

        # Cycle the AEN (ALS Enable) bit in the Enable Register (truly necessary?)
        # Enable Register (0x00)
        data    = [self.CMD + 0x00, 0x01]   # ALS Disable
        rbytes  = 1
        answ        = gglobs.elv.ELVaskDongle(self.addr, data, rbytes, name=self.name, info="Disable AEN")

        # Enable Register (0x00)
        data    = [self.CMD + 0x00, 0x03]   # ALS Enable
        rbytes  = 1
        answ        = gglobs.elv.ELVaskDongle(self.addr, data, rbytes, name=self.name, info="Enable AEN")

        start = time.time()
        time.sleep(intTime / 1000 * 1.5  ) # sleeping for integration time is almost
                                           # always enough to finish conversion
                                           # added 50%

        # Read the Status register until the AVALID bit (Bit #0 in Status) is set
        # Status Register (0x13)
        data    = [self.CMD + 0x13]
        rbytes  = 1
        answ        = gglobs.elv.ELVaskDongle(self.addr, data, rbytes, name=self.name, info="status", doPrint = True, end = "")
        if answ[0] & 0x01:
     #       util.ncprint("Data ready")
            pass
        else:
            ecprint("Data not ready", end="")
            ecprint(".", end="") # one dot for each call of status
            while True:
                if answ[0] & 0x01:
                    ncprint("ready after {:3.2f}sec".format(time.time() - start))
                    break
                data    = [self.CMD + 0x13] # Status is Register (0x13)
                rbytes  = 1
                answ        = gglobs.elv.ELVaskDongle(self.addr, data, rbytes, name=self.name, info="status", doPrint = False)
                ecprint(".", end="")

        # ALS Data Register (0x14 - 0x17)
        data   = [self.CMD + 0x14]
        rbytes = 4
        answ   = gglobs.elv.ELVaskDongle(self.addr, data, rbytes, name=self.name, info="Get data", end="")

        visraw = answ[0] | (answ[1] << 8)
        irraw  = answ[2] | (answ[3] << 8)

        # Results are validated for being a good approximation by this
        # normalization over all Gain factors
        vis    = visraw  / gainFct / intFct
        ir     = irraw   / gainFct / intFct

        return vis, ir, visraw, irraw, gainFct, intTime # need visraw for auto function in TSL2591getLumAuto


    def TSL2591getInfo(self):

        info = """{} (Category: {})
- DeviceID:   0x{:02X}
- Address:    0x{:02X}
""".format(self.name, "Ambient Light", self.subtype, self.addr)

        return info.split("\n")


    def TSL2591Reset(self):
        """Reset light sensor"""

        # System Reset
        # Control Register (0x01)
        # Important: the System Reset will NOT return an ACK! (observed on both ELV and IOW)
        data    = [self.CMD + 0x01, 0x80] # System Reset, AGAIN=00, ATIME=000
        rbytes  = 1
        #answ    = util.askDongle(self.dongle, self.addr, data, rbytes, name=self.name, info="System Reset")
        answ         = gglobs.elv.ELVaskDongle(self.addr, data, rbytes, name=self.name, info="System Reset")


    def TSL2591runAllFunctions(self):
        """ for TSL2591 sensor """

        # not running until the "answ    = util.askDongle(..." are exchanged for
        #                      " answ    = gglobs.elv.ELVaskDongle(..."
        return


    # Get Device Identification = 0x50  (= as subtype)
        # ID Register (0x12) (Bit 7:0)
        data    = [self.CMD + 0x12]
        rbytes  = 1
        answ    = util.askDongle(self.dongle, self.addr, data, rbytes, name=self.name, info="get ID")
        if answ[0] == self.subtype:
            print("Found Sensor TSL2591")
        else:
            print("Did NOT find Sensor TSL2591")
            #sys.exit()

    # Get package identification (PID)
        # PID Register (0x11) (Bit 5:4) (2 bits only!)
        data    = [self.CMD + 0x11]
        rbytes  = 1
        answ    = util.askDongle(self.dongle, self.addr, data, rbytes, name=self.name, info="PID read (B5:4)")
        pid     = answ[0] & 0b00110000
        if pid == self.PID:
            print("Package Identification '0b{:02b}' confirmed".format(pid), color = TGREEN)
        else:
            print("Package Identification '0b{:02b}' not as expected".format(pid))

    # System Reset
        # Control Register (0x01)
        # Important: the System Reset will NOT return an ACK! (observed on both ELV and IOW)
        data    = [self.CMD + 0x01, 0x80] # System Reset, AGAIN=00, ATIME=000
        rbytes  = 1
        answ    = util.askDongle(self.dongle, self.addr, data, rbytes, name=self.name, info="System Reset")

    # Enable measurement
        # Enable Register (0x00)
        # Power the device on, enable measurements
        # Bit#0 = 1: Power ON
        # Bit#1 = 1: ALS Enable
        # Register: 0b 0000 00 11  =0x03 :
        data    = [self.CMD + 0x00, 0x03]
        rbytes  = 1
        answ    = util.askDongle(self.dongle, self.addr, data, rbytes, name=self.name, info="Enable ALS+PON")

    # Get status
        # Status Register (0x13)
        data    = [self.CMD + 0x13]
        rbytes  = 1
        answ    = util.askDongle(self.dongle, self.addr, data, rbytes, name=self.name, info="status")

    # Get data
        # ALS Data Register (0x14 - 0x17)
        data    = [self.CMD + 0x14]
        rbytes  = 4
        answ    = util.askDongle(self.dongle, self.addr, data, rbytes, name=self.name, info="data")
        print("Data:", answ)

    # Set Med gain
        # Control Register (0x01)
        data    = [self.CMD + 0x01, 0x10]
        rbytes  = 1
        answ    = util.askDongle(self.dongle, self.addr, data, rbytes, name=self.name, info="Med gain")

    # Set High gain
        # Control Register (0x01)
        data    = [self.CMD + 0x01, 0x20]
        rbytes  = 1
        answ    = util.askDongle(self.dongle, self.addr, data, rbytes, name=self.name, info="High gain")

    # Set High gain + 300 ms integration
        # Control Register (0x01)
        data    = [self.CMD + 0x01, 0x22]
        rbytes  = 1
        answ    = util.askDongle(self.dongle, self.addr, data, rbytes, name=self.name, info="High gain+300ms")

    # Set Med gain + 300 ms integration
        # Control Register (0x01)
        data    = [self.CMD + 0x01, 0x12]
        rbytes  = 1
        answ    = util.askDongle(self.dongle, self.addr, data, rbytes, name=self.name, info="Med gain+300ms")

    # Set Med gain + 600 ms integration
        # Control Register (0x01)
        data    = [self.CMD + 0x01, 0x15]
        rbytes  = 1
        answ    = util.askDongle(self.dongle, self.addr, data, rbytes, name=self.name, info="Med gain+600ms")


def ncprint(*args, color = HILITECOLOR, end = "\n"):
    """Normal color print """

    print(color, end='')
    print(*args, end='')
    print(NORMALCOLOR, end=end)


def ecprint(*args, color = ERRORCOLOR, end="\n"):
    """Error color print """

    ncprint(*args, color = color, end=end)


